from pathlib import Path

import typer
import yaml
from pydantic import ValidationError
from rich import console as rich_console
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from jupyter_deploy import constants, fs_utils, manifest, variables_config


class NotADictError(ValueError):
    pass


class BaseProjectHandler:
    """Abstract class responsible for identifying the type of project.

    The current working directory MUST be a jupyter-deploy directory,
    otherwise this class will raise a typer.Exit().
    """

    def __init__(self) -> None:
        """Attempts to identify the engine associated with the project."""
        self._console: rich_console.Console | None = None
        self.project_path = Path.cwd()
        manifest_path = self.project_path / constants.MANIFEST_FILENAME

        try:
            project_manifest = retrieve_project_manifest(manifest_path)
        except FileNotFoundError as e:
            console = self.get_console()
            console.print(":x: The path does not correspond to a jupyter-deploy project.", style="bold red")
            console.line()
            console.print("Reason: could not find the jupyter-deploy manifest file.", style="red")
            console.print(f"Expected manifest file location: {manifest_path.absolute()}", style="red")
            console.line()
            console.print(
                "Change your working directory to a jupyter-deploy project or create one with [bold]jd init PATH[/].",
                style="red",
            )
            raise typer.Exit(1) from e
        except OSError as e:
            console = self.get_console()
            console.print(":x: Could not access the jupyter-deploy manifest.", style="bold red")
            console.line()
            console.print("Reason: OS error when reading the jupyter-deploy manifest file.", style="red")
            console.print(f"Manifest file location: {manifest_path.absolute()}", style="red")
            console.line()
            console.print(str(e))
            console.line()
            console.print("Verify your file system permissions and try again.", style="red")
            raise typer.Exit(1) from e
        except NotADictError as e:
            console = self.get_console()
            console.print(":x: The jupyter-deploy manifest is invalid.", style="bold red")
            console.line()
            console.print("Reason: expected the jupyter-deploy manifest file to parse as dict.", style="red")
            console.print(f"Attempted to read manifest file at: {manifest_path.absolute()}", style="red")
            raise typer.Exit(1) from e
        except (ParserError, ScannerError) as e:
            console = self.get_console()
            console.print(":x: The jupyter-deploy manifest is invalid.", style="bold red")
            console.line()
            console.print("Reason: cannot parse the jupyter-deploy manifest content as YAML.", style="red")
            console.print(f"Attempted to read manifest file at: {manifest_path.absolute()}", style="red")
            console.line()
            console.print(str(e))
            raise typer.Exit(1) from e
        except ValidationError as e:
            console = self.get_console()
            console.print(":x: The jupyter-deploy manifest is invalid.", style="bold red")
            console.line()
            console.print("Reason: the manifest file does not conform to the expected schema.", style="red")
            console.print(f"Attempted to read manifest file at: {manifest_path.absolute()}", style="red")
            console.line()
            console.print("Details:", style="red")
            for err in e.errors():
                console.print(err)
            raise typer.Exit(1) from e

        self.engine = project_manifest.get_engine()
        self.project_manifest = project_manifest

    def get_console(self) -> rich_console.Console:
        """Return the instance's rich console."""
        if self._console:
            return self._console
        self._console = rich_console.Console()
        return self._console


def retrieve_project_manifest(manifest_path: Path) -> manifest.JupyterDeployManifest:
    """Read the manifest file on disk, parse, validate and return it."""
    if not fs_utils.file_exists(manifest_path):
        raise FileNotFoundError("Missing jupyter-deploy manifest.")

    with open(manifest_path) as manifest_file:
        content = yaml.safe_load(manifest_file)

    if not isinstance(content, dict):
        raise NotADictError("Invalid manifest: jupyter-deploy manifest is not a dict.")

    return manifest.JupyterDeployManifest(**content)


def retrieve_project_manifest_if_available(project_path: Path) -> manifest.JupyterDeployManifest | None:
    """Attempts to read the manifest file on disk, parse, and return.

    Return None if the file is not found, cannot be parsed or fails any of the validation.
    """

    manifest_path = project_path / constants.MANIFEST_FILENAME
    try:
        return retrieve_project_manifest(manifest_path)
    except (FileNotFoundError, NotADictError, ValidationError) as e:
        print(e)
        return None


def retrieve_variables_config(variables_config_path: Path) -> variables_config.JupyterDeployVariablesConfig:
    """Read the variables confign file on disk, parse, validate and return it."""

    if not fs_utils.file_exists(variables_config_path):
        raise FileNotFoundError("Missing jupyter-deploy variables config file.")

    with open(variables_config_path) as variables_manifest_file:
        content = yaml.safe_load(variables_manifest_file)

    if not isinstance(content, dict):
        raise NotADictError("Invalid variables config file: jupyter-deploy variables config is not a dict.")

    return variables_config.JupyterDeployVariablesConfig(**content)



from enum import Enum


class EngineType(str, Enum):
    """Enum to list the types of deployment engine."""

    TERRAFORM = "terraform"

    @classmethod
    def from_string(cls, source_str: str) -> "EngineType":
        """Return the enum value, ignoring case.

        Raises:
            ValueError: If no matching enum value is found.
        """
        source_lower = source_str.lower()
        for source in cls:
            if source.value.lower() == source_lower:
                return source
        raise ValueError(f"No EngineType found for '{source_str}'")



import os
import select
import shutil
import subprocess
import sys
import threading
import time
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import IO


def check_executable_installation(
    executable_name: str, version_cmds: list[str] | None = None
) -> tuple[bool, str | None, str | None]:
    """Call which command on the package, return bool flag, version, error message."""
    if version_cmds is None:
        version_cmds = ["--version"]

    if shutil.which(executable_name) is None:
        return False, None, f"{executable_name} executable not found in system PATH"

    # Then try to run 'package --version' cmd
    try:
        cmd = [executable_name] + version_cmds
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=True,
        )
        # Extract version info from output
        version = result.stdout.strip().split("\n")[0]
        return True, version, None
    except FileNotFoundError:
        # This is a fallback in case shutil.which() returns a path but the file isn't actually executable
        return False, None, f"{executable_name} found in PATH, but executable not found."
    except subprocess.CalledProcessError as e:
        error_msg = e.stderr.strip() if e.stderr else "Unknown error"
        return False, None, error_msg
    except Exception as e:
        return False, None, f"{e}"


@contextmanager
def switch_dir(dir_path: Path | None) -> Generator:
    """Execute the inner function within a `cd` to the dir_path argument.

    If target_path is None, just execute the inner function.

    Raise:
        ValueError if the target_dir is not a valid path, or is not a directory.

    Usage:
        with project_dir("/path/to/dir"):
            # do something in that directory
    """

    if dir_path is None:
        yield
        return

    if not dir_path.exists():
        raise ValueError(f"Target path not found: {dir_path.absolute()}")
    elif not dir_path.is_dir():
        raise ValueError(f"Target path is not a directory: {dir_path.absolute()}")

    original_dir = Path(os.getcwd())
    try:
        os.chdir(dir_path)
        yield
    finally:
        os.chdir(original_dir)


def run_cmd_and_capture_output(cmds: list[str], exec_dir: Path | None = None) -> str:
    """Run command, returns output.

    Raises:
        CalledProcessError if return code is not 0
    """
    with switch_dir(exec_dir):
        result = subprocess.run(
            cmds,
            capture_output=True,
            text=True,
            check=True,
        )

    return result.stdout


def run_cmd_and_pipe_to_terminal(
    cmds: list[str], timeout_seconds: int | None = None, exec_dir: Path | None = None
) -> tuple[int, bool]:
    """Run command in a new process, pipe input in and output/error out to current.

    It will appear as though the command is being run in the current process.
    """

    class Timeout:
        _is_timedout = False

        @staticmethod
        def is_timedout() -> bool:
            return Timeout._is_timedout

        @staticmethod
        def set_timedout(is_timedout: bool) -> None:
            Timeout._is_timedout = is_timedout

    with switch_dir(exec_dir):
        p = subprocess.Popen(
            cmds,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            bufsize=0,
        )

        retcode: int | None = None
        timer: threading.Timer | None = None

        # Signal for when a prompt is likely waiting for input
        prompt_ready = threading.Event()
        stdout_active = threading.Event()

        # Buffer to store the stderr output while running
        stderr_buffer = []
        stderr_lock = threading.Lock()

        def timerout(p: subprocess.Popen) -> None:
            print(f"Command timed out after {timeout_seconds} second(s).")
            if timer:
                timer.cancel()
            Timeout.set_timedout(True)
            # terminating the subprocess may print a BrokenPipeError to stdout
            p.terminate()

        if timeout_seconds:
            timer = threading.Timer(timeout_seconds, timerout, args=[p])
            timer.start()

        # Create threads to handle input, output, and error streams concurrently
        def handle_stdin() -> None:
            try:
                while p.poll() is None:  # Continue as long as the process is running
                    # Wait for a prompt to appear
                    prompt_ready.wait(timeout=0.2)

                    # Add a small delay to ensure the full prompt is displayed
                    time.sleep(0.1)

                    # Clear the event for the next prompt
                    prompt_ready.clear()

                    try:
                        if sys.stdin.isatty():  # Check if stdin is a terminal
                            # Use non-blocking read for terminals
                            # TODO: check if this works on windows
                            rlist, _, _ = select.select([sys.stdin], [], [], 0.1)
                            if rlist and p.stdin:
                                line = sys.stdin.readline()
                                if not line:  # EOF (Ctrl+D)
                                    break
                                p.stdin.write(line)
                                p.stdin.flush()
                        elif p.stdin:
                            # For non-interactive input (e.g., piped input)
                            char = sys.stdin.read(1)
                            if not char:  # EOF
                                break
                            p.stdin.write(char)
                            p.stdin.flush()
                    except (OSError, BrokenPipeError):
                        break
            finally:
                if p.stdin:
                    p.stdin.close()

        def read_char_by_char(stream: IO[str]) -> None:
            """Read output character-by-character to handle prompts without newlines."""
            output_buffer = ""
            prompt_indicators = ["Enter a value: ", "?", ": "]  # Common prompt endings

            while True:
                char = stream.read(1)
                if not char:
                    break

                # Print the character immediately
                print(char, end="", flush=True)
                output_buffer += char

                # Check for common prompt indicators
                if any(output_buffer.endswith(indicator) for indicator in prompt_indicators):
                    prompt_ready.set()

                # Clear buffer after newline for efficiency
                if char == "\n":
                    output_buffer = ""
                    prompt_ready.set()  # Also check after each line

                # Keep a reasonable buffer size
                if len(output_buffer) > 100:
                    output_buffer = output_buffer[-50:]

            # Set the prompt_ready event one last time to unblock stdin thread
            prompt_ready.set()
            stream.close()

        def buffer_stderr(stream: IO[str]) -> None:
            """Buffer stderr output to print at the end."""
            while True:
                line = stream.readline()
                if not line:
                    break

                with stderr_lock:
                    stderr_buffer.append(line)

            stream.close()

        def handle_stdout() -> None:
            stdout_active.set()
            if p.stdout:
                read_char_by_char(p.stdout)
            stdout_active.clear()

        def handle_stderr() -> None:
            if p.stderr:
                buffer_stderr(p.stderr)

        # Start threads for I/O
        stdin_thread = threading.Thread(target=handle_stdin)
        stdout_thread = threading.Thread(target=handle_stdout)
        stderr_thread = threading.Thread(target=handle_stderr)

        # Set as daemon so it exits when main thread exits
        stdin_thread.daemon = True
        stdout_thread.daemon = True
        stderr_thread.daemon = True

        # Start output threads first
        stdout_thread.start()
        stderr_thread.start()
        stdin_thread.start()

        # Wait for process to complete
        retcode = p.wait()

        # Signal to unblock any waiting threads
        prompt_ready.set()

        # Wait for output threads to complete (with timeout to prevent hanging)
        stdout_thread.join(timeout=1)
        stderr_thread.join(timeout=1)
        stdin_thread.join(timeout=1)

        # Print the stderr output (if any) only after stdout is complete
        if stderr_buffer:
            for line in stderr_buffer:
                print(line, end="", flush=True)

        if timer:
            timer.cancel()

        return retcode, Timeout.is_timedout()


@contextmanager
def project_dir(dir: str | None) -> Generator:
    """Execute the inner function within a `cd` to the dir_path argument.

    If target_path is None, just execute the inner function.

    Raise:
        ValueError if the target_dir is not a valid path, or is not a directory.

    Usage:
        with project_dir("/path/to/dir"):
            # do something in that directory
    """
    if dir is None:
        yield
        return

    original_dir = Path(os.getcwd())
    target_path = Path(dir)

    if not target_path.exists():
        raise ValueError(f"Target path not found: {target_path}")
    elif not target_path.is_dir():
        raise ValueError(f"Target path is not a directory: {target_path}")

    try:
        os.chdir(target_path)
        yield
    finally:
        os.chdir(original_dir)



from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.engine.outdefs import StrTemplateOutputDefinition
from jupyter_deploy.provider.enum import ProviderType
from jupyter_deploy.provider.instruction_runner import InstructionRunner


class InstructionRunnerFactory:
    """Factory class to handle lower level imports of cloud provider specific dependencies.

    This ensures that the base jupyter-deploy does not depend on any cloud provider SDK.
    """

    _provider_runner_map: dict[ProviderType, InstructionRunner] = {}

    @staticmethod
    def get_provider_instruction_runner(api_name: str, outputs_handler: EngineOutputsHandler) -> InstructionRunner:
        """Return the instruction runner for the cloud provider.

        Raises:
            NotImplementedError if the provider is not recognized.
            ValueError if the provider runner requires declared values that are missing in manifest.
        """
        provider = ProviderType.from_api_name(api_name)
        if provider_runner := InstructionRunnerFactory._provider_runner_map.get(provider):
            return provider_runner

        if provider == ProviderType.AWS:
            aws_region_def = outputs_handler.get_declared_output_def("aws_region", StrTemplateOutputDefinition)

            # do NOT move import to top level
            from jupyter_deploy.provider.aws import aws_runner

            provider_runner = aws_runner.AwsApiRunner(region_name=aws_region_def.value)
            InstructionRunnerFactory._provider_runner_map[provider] = provider_runner
            return provider_runner

        raise NotImplementedError(f"No runner implementation for provider: {provider}")



"""Constants for Terraform operations."""

# Command constants
TF_INIT_CMD = ["terraform", "init"]
TF_PLAN_CMD = ["terraform", "plan"]
TF_APPLY_CMD = ["terraform", "apply"]
TF_DESTROY_CMD = ["terraform", "destroy"]
TF_OUTPUT_CMD = ["terraform", "output", "-json"]
TF_PARSE_PLAN_CMD = ["terraform", "show", "-json"]
TF_AUTO_APPROVE_CMD_OPTION = "-auto-approve"
TF_RM_FROM_STATE_CMD = ["terraform", "state", "rm"]

# Directory constants
TF_ENGINE_DIR = "engine"
TF_PRESETS_DIR = "presets"

# File constants
TF_DEFAULT_PLAN_FILENAME = "jdout-tfplan"
TF_RECORDED_VARS_FILENAME = "jdinputs.auto.tfvars"
TF_RECORDED_SECRETS_FILENAME = "jdinputs.secrets.auto.tfvars"
TF_CUSTOM_PRESET_FILENAME = "jdinputs.preset.override.tfvars"
TF_DESTROY_PRESET_FILENAME = "destroy.tfvars"
TF_VARIABLES_FILENAME = "variables.tf"
TF_OUTPUTS_FILENAME = "outputs.tf"


def get_preset_filename(preset_name: str = "all") -> str:
    """Return the full preset filename."""
    return f"defaults-{preset_name}.tfvars"



from typing import Any, Generic, TypeVar

from pydantic import BaseModel, ConfigDict

from jupyter_deploy.engine.outdefs import StrTemplateOutputDefinition, TemplateOutputDefinition
from jupyter_deploy.provider.resolved_clidefs import (
    IntResolvedCliParameter,
    ListStrResolvedCliParameter,
    ResolvedCliParameter,
    StrResolvedCliParameter,
)
from jupyter_deploy.provider.resolved_resultdefs import (
    IntResolvedInstructionResult,
    ListStrResolvedInstructionResult,
    ResolvedInstructionResult,
    StrResolvedInstructionResult,
)

S = TypeVar("S")


class ResolvedInstructionArgument(BaseModel, Generic[S]):
    """Wrapper class for a resolved value passed to instruction.

    The resolved value may come from the CLI arguments, project outputs,
    or results from other instructions.
    """

    model_config = ConfigDict(extra="allow")
    argument_name: str
    value: S


class StrResolvedInstructionArgument(ResolvedInstructionArgument[str]):
    """Wrapper class for a resolved value of type str passed to instruction."""

    pass


class IntResolvedInstructionArgument(ResolvedInstructionArgument[int]):
    """Wrapper class for a resolved value of type int passed to instruction."""

    pass


class ListStrResolvedInstructionArgument(ResolvedInstructionArgument[list[str]]):
    """Wrapper class for a resolved value of type list[str] passed to instruction."""

    pass


RIA = TypeVar("RIA", bound=ResolvedInstructionArgument)


def resolve_output_argdef(
    outdefs: dict[str, TemplateOutputDefinition], arg_name: str, source_key: str
) -> ResolvedInstructionArgument:
    """Instantiates the resolved argdef of the corresponding type.

    Raises:
        KeyError if the argument cannot be matched by source_key
        ValueError if the argument was not resolved
        NotImplementedError if there is no matching argtype class
    """

    if source_key not in outdefs:
        raise KeyError(f"Output name '{source_key}' not found in outputs: {list(outdefs.keys())}")
    outdef = outdefs[source_key]
    if isinstance(outdef, StrTemplateOutputDefinition):
        val = outdef.value
        if val is None:
            raise ValueError(f"Output name is not resolved: {source_key}")
        return StrResolvedInstructionArgument(argument_name=arg_name, value=val)

    raise NotImplementedError(f"No resolved argument class for type: {type(outdef).__name__}")


def resolve_result_argdef(
    resultdefs: dict[str, ResolvedInstructionResult], arg_name: str, source_key: str
) -> ResolvedInstructionArgument:
    """Instantiates the resolved argdef of the corresponding type.

    Raises:
        KeyError if the argument cannot be matched by source_key
        NotImplementedError if there is no matching argtype class
    """

    if source_key not in resultdefs:
        raise KeyError(f"Output name '{source_key}' not found in previous results: {list(resultdefs.keys())}")
    resultdef = resultdefs[source_key]
    if isinstance(resultdef, StrResolvedInstructionResult):
        return StrResolvedInstructionArgument(argument_name=arg_name, value=resultdef.value)
    elif isinstance(resultdef, IntResolvedInstructionResult):
        return IntResolvedInstructionArgument(argument_name=arg_name, value=resultdef.value)
    elif isinstance(resultdef, ListStrResolvedInstructionResult):
        return ListStrResolvedInstructionArgument(argument_name=arg_name, value=resultdef.value)

    raise NotImplementedError(f"No resolved argument class for type: {type(resultdef).__name__}")


def resolve_cliparam_argdef(
    paramdefs: dict[str, ResolvedCliParameter], arg_name: str, source_key: str
) -> ResolvedInstructionArgument:
    """Instantiates the resolved argdef of the corresponding type.

    Raises:
        KeyError if the argument cannot be matched by source_key
        NotImplementedError if there is no matching argtype class
    """

    if source_key not in paramdefs:
        raise KeyError(f"Output name '{source_key}' not found in CLI parameters: {list(paramdefs.keys())}")
    paramdef = paramdefs[source_key]
    if isinstance(paramdef, StrResolvedCliParameter):
        return StrResolvedInstructionArgument(argument_name=arg_name, value=paramdef.value)
    elif isinstance(paramdef, IntResolvedCliParameter):
        return IntResolvedInstructionArgument(argument_name=arg_name, value=paramdef.value)
    elif isinstance(paramdef, ListStrResolvedCliParameter):
        return ListStrResolvedInstructionArgument(argument_name=arg_name, value=paramdef.value)

    raise NotImplementedError(f"No resolved argument class for type: {type(paramdef).__name__}")


def require_arg(resolved_args: dict[str, ResolvedInstructionArgument], arg_name: str, arg_type: type[RIA]) -> RIA:
    """Returns the resolved arg of the expected type from the args dict.

    Args:
        resolved_args: A dictionary of resolved arguments.
        arg_name: The name of the argument to retrieve.
        arg_type: The expected type of the argument, a subclass of ResolvedInstructionArgument.

    Raises:
        KeyError: If arg_name is not in resolved_args.
        TypeError: If the argument is not of the expected type.
    """
    if arg_name not in resolved_args:
        raise KeyError(f"Required argument '{arg_name}' not found in resolved arguments")

    arg = resolved_args[arg_name]

    if not isinstance(arg, arg_type):
        raise TypeError(f"Expected argument '{arg_name}' to be of type {arg_type.__name__}, got {type(arg).__name__}")

    return arg


def retrieve_optional_arg(
    resolved_args: dict[str, ResolvedInstructionArgument], arg_name: str, arg_type: type[RIA], default_value: Any
) -> RIA:
    """Returns the resolved arg of the expected type from the args dict, or None if not found.

    Args:
        resolved_args: A dictionary of resolved arguments.
        arg_name: The name of the argument to retrieve.
        arg_type: The expected type of the argument, a subclass of ResolvedInstructionArgument.

    Raises:
        TypeError: If the argument is not of the expected type.
    """
    if arg_name not in resolved_args:
        return arg_type(argument_name=arg_name, value=default_value)

    arg = resolved_args[arg_name]

    if not isinstance(arg, arg_type):
        raise TypeError(f"Expected argument '{arg_name}' to be of type {arg_type.__name__}, got {type(arg).__name__}")

    return arg



from enum import Enum


class ProviderType(str, Enum):
    """Enum to list the types of cloud providers."""

    AWS = "aws"

    @classmethod
    def from_api_name(cls, api_name: str) -> "ProviderType":
        """Return the enum value, ignoring case.

        Raises:
            ValueError: If no matching enum value is found.
        """
        parts = api_name.split(".")
        source_lower = parts[0].lower()
        for source in cls:
            if source.value.lower() == source_lower:
                return source
        raise ValueError(f"No ProviderType found for api name: {api_name}")



"""Template utilities for jupyter-deploy."""

import importlib.metadata
import logging
from pathlib import Path

from jupyter_deploy.engine.enum import EngineType

logger = logging.getLogger(__name__)

TEMPLATE_ENTRY_POINTS = {engine_type: f"jupyter_deploy.{engine_type.value}_templates" for engine_type in EngineType}

TEMPLATES: dict[str, dict[str, Path]] = {engine: {} for engine in TEMPLATE_ENTRY_POINTS}


def get_templates(engine: EngineType) -> dict[str, Path]:
    """Get all registered templates for a specific engine from entry points.

    Args:
        engine: The engine type (e.g., "terraform")

    Returns:
        Dict[str, Path]: A dictionary mapping template names to their paths.
    """
    templates: dict[str, Path] = {}

    if engine not in TEMPLATE_ENTRY_POINTS:
        logger.warning(f"No entry point defined for engine: {engine}")
        return templates

    entry_point_group = TEMPLATE_ENTRY_POINTS[engine]

    for entry_point in importlib.metadata.entry_points(group=entry_point_group):
        template_path = entry_point.load()
        if isinstance(template_path, Path) and template_path.exists():
            # Convert entry point name to template format, ex. aws_ec2_some-name -> aws:ec2:some_name
            template_name = entry_point.name.replace("_", ":")
            templates[template_name] = template_path
            logger.debug(f"Loaded {engine} template {template_name} from {template_path}")
        else:
            logger.warning(f"Template path for {entry_point.name} is not a valid Path or does not exist")

    return templates


for engine in TEMPLATE_ENTRY_POINTS:
    TEMPLATES[engine] = get_templates(engine)



import unittest
from unittest.mock import Mock, patch

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.handlers.project.show_handler import ShowHandler
from jupyter_deploy.manifest import JupyterDeployManifestV1


class TestShowHandler(unittest.TestCase):
    def get_mock_manifest(self) -> JupyterDeployManifestV1:
        """Create a mock manifest."""
        return JupyterDeployManifestV1(
            **{  # type: ignore
                "schema_version": 1,
                "template": {
                    "name": "tf-aws-ec2-base",
                    "engine": "terraform",
                    "version": "1.0.0",
                },
            }
        )

    def get_mock_console_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mocked rich console instance."""
        mock_console = Mock()
        mock_print = Mock()
        mock_line = Mock()
        mock_console.print = mock_print
        mock_console.line = mock_line
        return mock_console, {"print": mock_print, "line": mock_line}

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    def test_init_terraform(self, mock_retrieve_manifest: Mock) -> None:
        mock_manifest = self.get_mock_manifest()
        mock_retrieve_manifest.return_value = mock_manifest
        handler = ShowHandler()

        self.assertIsNotNone(handler._outputs_handler)
        self.assertIsNotNone(handler._variables_handler)
        self.assertEqual(handler.engine, EngineType.TERRAFORM)
        self.assertEqual(handler.project_manifest, mock_manifest)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("rich.console.Console")
    def test_show_project_basic_info(self, mock_console_cls: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.get_mock_manifest()
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console
        handler = ShowHandler()

        handler._show_project_basic_info()

        # Verify console.print was called (header + table)
        self.assertEqual(mock_console_fns["print"].call_count, 2)
        self.assertEqual(mock_console_fns["line"].call_count, 2)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("rich.console.Console")
    def test_show_project_outputs_no_outputs(self, mock_console_cls: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.get_mock_manifest()
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console
        handler = ShowHandler()

        with patch.object(handler._outputs_handler, "get_full_project_outputs", return_value={}) as _:
            handler._show_project_outputs()

        # Check that both lines are printed with correct styles
        mock_console_fns["print"].assert_any_call(":warning: No outputs available.", style="yellow")
        mock_console_fns["print"].assert_any_call(
            "This is normal if the project has not been deployed yet.", style="yellow"
        )
        self.assertEqual(mock_console_fns["print"].call_count, 2)
        mock_console_fns["line"].assert_called_once()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("rich.console.Console")
    def test_show_project_outputs_with_outputs(self, mock_console_cls: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.get_mock_manifest()
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console
        handler = ShowHandler()

        mock_output = Mock()
        mock_output.value = "https://example.com"
        mock_output.description = "Jupyter URL"
        mock_outputs = {"jupyter_url": mock_output}

        with patch.object(handler._outputs_handler, "get_full_project_outputs", return_value=mock_outputs) as _:
            handler._show_project_outputs()

        # Verify console.print was called (header + table)
        self.assertEqual(mock_console_fns["print"].call_count, 2)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("rich.console.Console")
    def test_show_project_outputs_exception(self, mock_console_cls: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.get_mock_manifest()
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console
        handler = ShowHandler()

        with patch.object(handler._outputs_handler, "get_full_project_outputs", side_effect=Exception("Test error")):
            handler._show_project_outputs()

        # Verify error handling
        mock_console_fns["print"].assert_any_call(":x: Could not retrieve outputs: Test error", style="red")
        mock_console_fns["line"].assert_called_once()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("rich.console.Console")
    def test_show_project_variables_no_variables(self, mock_console_cls: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.get_mock_manifest()
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console
        handler = ShowHandler()

        with (
            patch.object(handler._variables_handler, "get_template_variables", return_value={}),
            patch.object(handler._variables_handler, "sync_engine_varfiles_with_project_variables_config"),
        ):
            handler._show_project_variables()

        mock_console_fns["print"].assert_called_once_with(":x: No variables available.", style="red")

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("rich.console.Console")
    def test_show_project_variables_with_variables(self, mock_console_cls: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.get_mock_manifest()
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console
        handler = ShowHandler()

        # Create mock variables
        mock_normal_var = Mock()
        mock_normal_var.get_cli_description = Mock(return_value="Normal variable")
        mock_normal_var.sensitive = False
        mock_normal_var.assigned_value = "value1"

        mock_sensitive_var = Mock()
        mock_sensitive_var.get_cli_description = Mock(return_value="Sensitive variable")
        mock_sensitive_var.sensitive = True
        mock_sensitive_var.assigned_value = "secret"

        mock_vars = {"normal_var": mock_normal_var, "sensitive_var": mock_sensitive_var}

        with (
            patch.object(handler._variables_handler, "get_template_variables", return_value=mock_vars),
            patch.object(handler._variables_handler, "sync_engine_varfiles_with_project_variables_config"),
        ):
            handler._show_project_variables()

        # Verify console.print was called (header + table)
        self.assertEqual(mock_console_fns["print"].call_count, 2)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    def test_sensitive_variables_are_masked(self, mock_retrieve_manifest: Mock) -> None:
        """Test that sensitive variables are properly masked in the output."""
        mock_retrieve_manifest.return_value = self.get_mock_manifest()
        handler = ShowHandler()

        # Create mock variables with one sensitive and one normal variable
        mock_normal_var = Mock()
        mock_normal_var.get_cli_description = Mock(return_value="Normal variable")
        mock_normal_var.sensitive = False
        mock_normal_var.assigned_value = "visible_value"

        mock_sensitive_var = Mock()
        mock_sensitive_var.get_cli_description = Mock(return_value="Sensitive variable")
        mock_sensitive_var.sensitive = True
        mock_sensitive_var.assigned_value = "secret_value_should_not_be_visible"

        mock_vars = {"normal_var": mock_normal_var, "sensitive_var": mock_sensitive_var}

        # Mock the actual Table.add_row method to capture the values being added
        rows_added: list[tuple[str, ...]] = []

        def capture_add_row(*args: str, **kwargs: object) -> None:
            rows_added.append(args)
            return None

        with (
            patch.object(handler._variables_handler, "get_template_variables", return_value=mock_vars),
            patch.object(handler._variables_handler, "sync_engine_varfiles_with_project_variables_config"),
            patch("rich.table.Table.add_row", side_effect=capture_add_row),
        ):
            handler._show_project_variables()

        # Find rows for our variables
        normal_var_row: tuple[str, ...] | None = None
        sensitive_var_row: tuple[str, ...] | None = None

        for row in rows_added:
            if len(row) >= 3:  # Ensure row has enough elements
                if row[0] == "normal_var":
                    normal_var_row = row
                elif row[0] == "sensitive_var":
                    sensitive_var_row = row

        # Verify both variables were added to the table
        self.assertIsNotNone(normal_var_row, "Normal variable row not found in table")
        self.assertIsNotNone(sensitive_var_row, "Sensitive variable row not found in table")

        # Verify normal variable shows its actual value
        assert normal_var_row is not None  # For mypy
        self.assertEqual(normal_var_row[1], "visible_value", "Normal variable value should be visible")

        # Verify sensitive variable value is masked
        assert sensitive_var_row is not None  # For mypy
        self.assertEqual(sensitive_var_row[1], "****", "Sensitive variable should be masked with asterisks")
        self.assertNotEqual(
            sensitive_var_row[1],
            "secret_value_should_not_be_visible",
            "Sensitive variable value should not be displayed",
        )

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("rich.console.Console")
    def test_show_project_variables_exception(self, mock_console_cls: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.get_mock_manifest()
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console
        handler = ShowHandler()

        with patch.object(
            handler._variables_handler,
            "sync_engine_varfiles_with_project_variables_config",
            side_effect=Exception("Test error"),
        ):
            handler._show_project_variables()

        # Verify error handling
        mock_console_fns["print"].assert_called_once_with(":x: Could not retrieve variables: Test error", style="red")

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("rich.console.Console")
    def test_show_project_info_default(self, mock_console_cls: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.get_mock_manifest()
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console
        handler = ShowHandler()

        with (
            patch.object(handler, "_show_project_basic_info") as mock_basic,
            patch.object(handler, "_show_project_outputs") as mock_outputs,
            patch.object(handler, "_show_project_variables") as mock_variables,
        ):
            handler.show_project_info()
            mock_basic.assert_called_once()
            mock_outputs.assert_called_once()
            mock_variables.assert_called_once()

        self.assertEqual(mock_console_fns["line"].call_count, 3)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("rich.console.Console")
    def test_show_project_info_with_flags(self, mock_console_cls: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.get_mock_manifest()
        mock_console, _ = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console
        handler = ShowHandler()

        with (
            patch.object(handler, "_show_project_basic_info") as mock_basic,
            patch.object(handler, "_show_project_outputs") as mock_outputs,
            patch.object(handler, "_show_project_variables") as mock_variables,
        ):
            # Test with only info flag
            handler.show_project_info(show_info=True, show_outputs=False, show_variables=False)
            mock_basic.assert_called_once()
            mock_outputs.assert_not_called()
            mock_variables.assert_not_called()
            mock_basic.reset_mock()

            # Test with only outputs flag
            handler.show_project_info(show_info=False, show_outputs=True, show_variables=False)
            mock_basic.assert_not_called()
            mock_outputs.assert_called_once()
            mock_variables.assert_not_called()
            mock_outputs.reset_mock()

            # Test with only variables flag
            handler.show_project_info(show_info=False, show_outputs=False, show_variables=True)
            mock_basic.assert_not_called()
            mock_outputs.assert_not_called()
            mock_variables.assert_called_once()
            mock_variables.reset_mock()

            # Test with multiple flags
            handler.show_project_info(show_info=True, show_outputs=True, show_variables=False)
            mock_basic.assert_called_once()
            mock_outputs.assert_called_once()
            mock_variables.assert_not_called()



"""Terraform implementation of the `config` handler."""

from pathlib import Path
from subprocess import CalledProcessError
from typing import Any

from pydantic import ValidationError
from rich import console as rich_console

from jupyter_deploy import cmd_utils, fs_utils, verify_utils
from jupyter_deploy.engine.engine_config import EngineConfigHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_plan, tf_vardefs, tf_variables
from jupyter_deploy.engine.terraform.tf_constants import (
    TF_DEFAULT_PLAN_FILENAME,
    TF_ENGINE_DIR,
    TF_INIT_CMD,
    TF_PARSE_PLAN_CMD,
    TF_PLAN_CMD,
    TF_PRESETS_DIR,
    get_preset_filename,
)
from jupyter_deploy.engine.vardefs import TemplateVariableDefinition
from jupyter_deploy.manifest import JupyterDeployManifest


class TerraformConfigHandler(EngineConfigHandler):
    """Config handler implementation for terraform projects."""

    def __init__(
        self,
        project_path: Path,
        project_manifest: JupyterDeployManifest,
        output_filename: str | None = None,
    ) -> None:
        variables_handler = tf_variables.TerraformVariablesHandler(
            project_path=project_path, project_manifest=project_manifest
        )
        super().__init__(
            project_path=project_path,
            project_manifest=project_manifest,
            variables_handler=variables_handler,
            engine=EngineType.TERRAFORM,
        )
        self.engine_dir_path = project_path / TF_ENGINE_DIR
        self.plan_out_path = self.engine_dir_path / (output_filename or TF_DEFAULT_PLAN_FILENAME)

        # use a different name from parent attribute to not confuse mypy
        self.tf_variables_handler = variables_handler

    def _get_preset_path(self, preset_name: str) -> Path:
        return self.engine_dir_path / TF_PRESETS_DIR / get_preset_filename(preset_name)

    def has_recorded_variables(self) -> bool:
        file_path = self.tf_variables_handler.get_recorded_variables_filepath()
        return fs_utils.file_exists(file_path=file_path)

    def verify_preset_exists(self, preset_name: str) -> bool:
        file_path = self._get_preset_path(preset_name)
        return fs_utils.file_exists(file_path=file_path)

    def list_presets(self) -> list[str]:
        presets = ["none"]

        # Get all files matching the pattern
        matching_filenames = fs_utils.find_matching_filenames(
            dir_path=self.engine_dir_path / TF_PRESETS_DIR,
            file_pattern="defaults-*.tfvars",
        )
        presets.extend([n[len("defaults-") : -len(".tfvars")] for n in matching_filenames])
        return sorted(presets)

    def verify_requirements(self) -> bool:
        requirements = self.project_manifest.get_requirements()
        all_installed = verify_utils.verify_tools_installation(requirements)
        return all_installed

    def reset_recorded_variables(self) -> None:
        self.variables_handler.reset_recorded_variables()

    def reset_recorded_secrets(self) -> None:
        self.variables_handler.reset_recorded_secrets()

    def configure(
        self, preset_name: str | None = None, variable_overrides: dict[str, TemplateVariableDefinition] | None = None
    ) -> bool:
        console = rich_console.Console()

        # 1/ run terraform init.
        # Note that it is safe to run several times, see ``terraform init --help``:
        # ``init`` command is always safe to run multiple times. Though subsequent runs
        # may give errors, this command will never delete your configuration or
        # state.
        init_retcode, init_timed_out = cmd_utils.run_cmd_and_pipe_to_terminal(
            TF_INIT_CMD.copy(),
            exec_dir=self.engine_dir_path,
        )
        if init_retcode != 0 or init_timed_out:
            console.print(":x: Error initializing Terraform project.", style="red")
            return False

        # 2/ prepare to run terraform plan and save output with ``terraform plan PATH``
        plan_cmds = TF_PLAN_CMD.copy()

        # 2.1/ output plan to disk
        plan_cmds.append(f"-out={self.plan_out_path.absolute()}")

        # 2.2/ sync variables.yaml -> tfvars and create the .tfvars files if necessary
        self.variables_handler.sync_engine_varfiles_with_project_variables_config()

        # 2.3/ using preset
        if preset_name:
            # here we assume the preset path was verified earlier
            base_preset_path = self._get_preset_path(preset_name)

            # if a user i/ runs `jd init`, ii/ set values in variables.yaml,
            # iii/ calls `jd config`, then the `jdinputs.auto.tfvars` file
            # then the preset values may take precedence over the values specified
            # in variables.yaml, which is not desirable.
            filtered_preset_path = self.tf_variables_handler.create_filtered_preset_file(base_preset_path)
            plan_cmds.append(f"-var-file={filtered_preset_path.absolute()}")

        # 2.4/ pass variable overrides
        if variable_overrides:
            for var_def in variable_overrides.values():
                var_option = tf_vardefs.to_tf_var_option(var_def)
                plan_cmds.extend(var_option)

        # 2.5/ call terraform plan
        plan_retcode, plan_timed_out = cmd_utils.run_cmd_and_pipe_to_terminal(plan_cmds, exec_dir=self.engine_dir_path)

        if plan_retcode != 0 or plan_timed_out:
            console.line()
            console.print(":x: Error generating Terraform plan.", style="red")

        # on successful plan generation, terraform prints out where the plan is saved,
        # hence no need to print it again.
        return plan_retcode == 0 and not plan_timed_out

    def record(self, record_vars: bool = False, record_secrets: bool = False) -> None:
        if not record_vars and not record_secrets:
            return

        console = rich_console.Console()
        cmds = TF_PARSE_PLAN_CMD + [f"{self.plan_out_path.absolute()}"]

        try:
            plan_content_str = cmd_utils.run_cmd_and_capture_output(cmds, exec_dir=self.engine_dir_path)
        except CalledProcessError as e:
            console.print(f":x: Failed to retrieve plan at: {self.plan_out_path.name}")
            console.print(e.stdout)
            console.print(e.stderr)
            return

        try:
            variables, secrets = tf_plan.extract_variables_from_json_plan(plan_content_str)
        except (ValueError, ValidationError):  # noqa: B904
            console.print(f":x: invalid plan at: {self.plan_out_path.name}")
            return

        vardefs: dict[str, Any] = {}

        if record_vars:
            vars_file_path = self.tf_variables_handler.get_recorded_variables_filepath()
            vars_file_lines = ["# generated by jupyter-deploy config command\n"]
            vars_file_lines.extend(tf_plan.format_plan_variables(variables))
            fs_utils.write_inline_file_content(vars_file_path, vars_file_lines)
            console.print(f":floppy_disk: Recorded variables: {vars_file_path.name}")

            vardefs.update({k: v.value for k, v in variables.items()})

        if record_secrets:
            secrets_file_path = self.tf_variables_handler.get_recorded_secrets_filepath()
            secrets_file_lines = ["# generated by jupyter-deploy config command\n"]
            secrets_file_lines.append("# do NOT commit this file\n")
            secrets_file_lines.extend(tf_plan.format_plan_variables(secrets))
            fs_utils.write_inline_file_content(secrets_file_path, secrets_file_lines)
            console.print(f":floppy_disk: Recorded secrets: {secrets_file_path.name}")
            console.print(f":warning: Do [bold]not[/] commit the secret file: {secrets_file_path.name}", style="yellow")

            vardefs.update({k: v.value for k, v in secrets.items()})

        if record_vars or record_secrets:
            self.variables_handler.sync_project_variables_config(vardefs)
            variables_config_path = self.variables_handler.get_variables_config_path()
            console.line()
            console.print(f":floppy_disk: Recorded variables config: {variables_config_path.name}")



from mypy_boto3_ssm import SSMClient
from mypy_boto3_ssm.type_defs import DescribeInstanceInformationRequestTypeDef, InstanceInformationTypeDef


def describe_instance_information(ssm_client: SSMClient, instance_id: str) -> InstanceInformationTypeDef:
    """Call SSM:DescribeInstanceInformation, return the result."""

    request: DescribeInstanceInformationRequestTypeDef = {"Filters": [{"Key": "InstanceIds", "Values": [instance_id]}]}
    response = ssm_client.describe_instance_information(**request)
    information_list = response["InstanceInformationList"]

    if not information_list:
        raise ValueError("SSM:DescribeInstanceInformation returned an empty list")

    return information_list[0]



from typing import Annotated

import typer

from jupyter_deploy import cmd_utils
from jupyter_deploy.handlers.access import team_handler

teams_app = typer.Typer(
    help=("""Control access to your jupyter app at team level."""),
    no_args_is_help=True,
)


@teams_app.command()
def add(
    teams: Annotated[list[str], typer.Argument(help="Names of the teams to add to the allowlist.")],
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project."),
    ] = None,
) -> None:
    """Add team(s) to the list authorized to access the Jupyter app.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = team_handler.TeamsHandler()
        handler.add_teams(teams)


@teams_app.command()
def remove(
    teams: Annotated[list[str], typer.Argument(help="Names of the teams to remove from the allowlist.")],
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project."),
    ] = None,
) -> None:
    """Remove team(s) from the list authorized to access the Jupyter app.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = team_handler.TeamsHandler()
        handler.remove_teams(teams)


@teams_app.command()
def set(
    teams: Annotated[list[str], typer.Argument(help="Names of the teams to allowlist.")],
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project."),
    ] = None,
) -> None:
    """Set the list of team(s) authorized to access the Jupyter app.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = team_handler.TeamsHandler()
        handler.set_teams(teams)


# use a cmd alias because mypy shows an 'valid-type' error if we just call the method 'list'
@teams_app.command("list")
def list_teams(
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project."),
    ] = None,
) -> None:
    """Show the name(s) of the team(s) authorized to access the Jupyter app.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = team_handler.TeamsHandler()
        teams = handler.list_teams()
        console = handler.get_console()

        if teams:
            console.print(f"Allowlisted teams: [bold cyan]{', '.join(teams)}[/]")
        else:
            console.print("Allowlisted teams: [bold cyan]None[/]")



from enum import Enum


class AWSInfrastructureType(str, Enum):
    """Enum to list the types of AWS infrastructure."""

    EC2 = "ec2"


# Use the AWS infrastructure type as the default infrastructure type for now
# When more provider-specific infrastructure types are added, we can update this
# to use a Union type or create a new enum that includes all values from all provider-specific enums
InfrastructureType = AWSInfrastructureType



from typing import Annotated

import typer

from jupyter_deploy import cmd_utils
from jupyter_deploy.handlers.access import organization_handler

organization_app = typer.Typer(
    help=("""Control access to your jupyter app at the organization level."""),
    no_args_is_help=True,
)


@organization_app.command()
def set(
    organization: Annotated[str, typer.Argument(help="Name of the organization to allowlist.")],
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project."),
    ] = None,
) -> None:
    """Allowlist an organization to access the Jupyter app.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = organization_handler.OrganizationHandler()
        handler.set_organization(organization)


@organization_app.command()
def unset(
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project."),
    ] = None,
) -> None:
    """Remove organization-based access to the Jupyter app.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = organization_handler.OrganizationHandler()
        handler.unset_organization()


@organization_app.command()
def get(
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project."),
    ] = None,
) -> None:
    """Show the name of the organization authorized to access the Jupyter app.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = organization_handler.OrganizationHandler()
        organization = handler.get_organization()
        console = handler.get_console()

        if organization:
            console.print(f"Allowlisted organization: [bold cyan]{organization}[/]")
        else:
            console.print("Allowlisted organization: [bold cyan]None[/]")



import time

import botocore
import botocore.exceptions
from mypy_boto3_ssm.client import SSMClient
from mypy_boto3_ssm.literals import CommandInvocationStatusType
from mypy_boto3_ssm.type_defs import GetCommandInvocationResultTypeDef, SendCommandRequestTypeDef

TERMINAL_COMMAND_STATUS: list[CommandInvocationStatusType] = ["Cancelled", "Failed", "Success", "TimedOut"]


def is_terminal_command_invocation_status(command_status: CommandInvocationStatusType) -> bool:
    """Return True for terminal status, False otherwise."""
    return command_status in TERMINAL_COMMAND_STATUS


def poll_command(
    client: SSMClient,
    command_id: str,
    instance_id: str,
    poll_interval_seconds: int = 2,
    wait_on_invocation_does_not_exist: int = 2,
) -> GetCommandInvocationResultTypeDef:
    """Call SSM:GetCommandExecution until terminal state, return API response.

    The first call may fail as SSM takes 1-2 seconds to register a newly-sent command.
    This methid retries the error after a short wait.
    """
    try:
        result = client.get_command_invocation(CommandId=command_id, InstanceId=instance_id)
        status = result["Status"]
    except botocore.exceptions.ClientError as e:
        if e.response.get("Error", {}).get("Code") == "InvocationDoesNotExist":
            time.sleep(wait_on_invocation_does_not_exist)
            result = client.get_command_invocation(CommandId=command_id, InstanceId=instance_id)
            status = result["Status"]
        else:
            raise

    # note: this cannot be an infinite loop: commands have a time-out.
    while not is_terminal_command_invocation_status(status):
        time.sleep(poll_interval_seconds)
        result = client.get_command_invocation(CommandId=command_id, InstanceId=instance_id)
        status = result["Status"]

    return result


def send_cmd_to_one_instance_and_wait_sync(
    client: SSMClient,
    document_name: str,
    instance_id: str,
    timeout_seconds: int = 30,
    wait_after_send_seconds: int = 2,
    **parameters: list[str],
) -> GetCommandInvocationResultTypeDef:
    """Send the command, poll execution, return execution response."""
    request: SendCommandRequestTypeDef = {
        "DocumentName": document_name,
        "InstanceIds": [instance_id],
        "TimeoutSeconds": timeout_seconds,
    }
    if parameters:
        request["Parameters"] = parameters

    send_command_result = client.send_command(**request)

    command_id = send_command_result["Command"].get("CommandId")
    if not command_id:
        raise RuntimeError("Command ID could not be retrieved.")

    # give SSM time to register the command
    if wait_after_send_seconds > 0:
        time.sleep(wait_after_send_seconds)

    terminal_command_execution_response = poll_command(client, command_id=command_id, instance_id=instance_id)
    return terminal_command_execution_response



import unittest
from unittest.mock import Mock, patch

from typer.testing import CliRunner

from jupyter_deploy.cli.organization_app import organization_app


class TestOrganizationApp(unittest.TestCase):
    """Test cases for the organization_app module."""

    def test_help_command(self) -> None:
        """Test the help command."""
        self.assertTrue(len(organization_app.info.help or "") > 0, "help should not be empty")

        runner = CliRunner()
        result = runner.invoke(organization_app, ["--help"])

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(result.stdout.index("set") > 0)
        self.assertTrue(result.stdout.index("unset") > 0)
        self.assertTrue(result.stdout.index("get") > 0)

    def test_no_arg_defaults_to_help(self) -> None:
        """Test that running the app with no arguments shows help."""
        runner = CliRunner()
        result = runner.invoke(organization_app, [])

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(len(result.stdout) > 0)


class TestOrganizationSetCmd(unittest.TestCase):
    def get_mock_organization_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock organization handler."""
        mock_organization_handler = Mock()
        mock_set_organization = Mock()
        mock_get_console = Mock()

        mock_organization_handler.set_organization = mock_set_organization
        mock_organization_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_organization_handler, {
            "set_organization": mock_set_organization,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_organization_handler_and_calls_set_organization(
        self, mock_project_dir: Mock, mock_organization_handler_class: Mock
    ) -> None:
        """Test that set command instantiates OrganizationHandler and calls set_organization."""
        # Setup
        mock_organization_handler, mock_handler_fns = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["set", "org-name"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_organization_handler_class.assert_called_once()
        mock_handler_fns["set_organization"].assert_called_once_with("org-name")

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(
        self, mock_project_dir: Mock, mock_organization_handler_class: Mock
    ) -> None:
        """Test that set command switches directory when a project path is provided."""
        # Setup
        mock_organization_handler, _ = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["set", "org-name", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_organization_handler_set_organization_raises(
        self, mock_project_dir: Mock, mock_organization_handler_class: Mock
    ) -> None:
        """Test that set command propagates exceptions from set_organization."""
        # Setup
        mock_organization_handler, mock_handler_fns = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler
        mock_handler_fns["set_organization"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["set", "org-name"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)


class TestOrganizationUnsetCmd(unittest.TestCase):
    def get_mock_organization_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock organization handler."""
        mock_organization_handler = Mock()
        mock_unset_organization = Mock()
        mock_get_console = Mock()

        mock_organization_handler.unset_organization = mock_unset_organization
        mock_organization_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_organization_handler, {
            "unset_organization": mock_unset_organization,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_organization_handler_and_calls_unset_organization(
        self, mock_project_dir: Mock, mock_organization_handler_class: Mock
    ) -> None:
        """Test that unset command instantiates OrganizationHandler and calls unset_organization."""
        # Setup
        mock_organization_handler, mock_handler_fns = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["unset"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_organization_handler_class.assert_called_once()
        mock_handler_fns["unset_organization"].assert_called_once()

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(
        self, mock_project_dir: Mock, mock_organization_handler_class: Mock
    ) -> None:
        """Test that unset command switches directory when a project path is provided."""
        # Setup
        mock_organization_handler, _ = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["unset", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_organization_handler_unset_organization_raises(
        self, mock_project_dir: Mock, mock_organization_handler_class: Mock
    ) -> None:
        """Test that unset command propagates exceptions from unset_organization."""
        # Setup
        mock_organization_handler, mock_handler_fns = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler
        mock_handler_fns["unset_organization"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["unset"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)


class TestOrganizationGetCmd(unittest.TestCase):
    def get_mock_organization_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock organization handler."""
        mock_organization_handler = Mock()
        mock_get_organization = Mock()
        mock_get_console = Mock()

        mock_organization_handler.get_organization = mock_get_organization
        mock_organization_handler.get_console = mock_get_console

        mock_get_organization.return_value = "test-org"
        mock_get_console.return_value = Mock()

        return mock_organization_handler, {
            "get_organization": mock_get_organization,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_organization_handler_and_calls_get_organization(
        self, mock_project_dir: Mock, mock_organization_handler_class: Mock
    ) -> None:
        """Test that get command instantiates OrganizationHandler and calls get_organization."""
        # Setup
        mock_organization_handler, mock_handler_fns = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["get"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_organization_handler_class.assert_called_once()
        mock_handler_fns["get_organization"].assert_called_once()

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_uses_handler_console_to_print_organization(
        self, mock_project_dir: Mock, mock_organization_handler_class: Mock
    ) -> None:
        """Test that get command uses the handler's console to print the organization."""
        # Setup
        mock_organization_handler, mock_handler_fns = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler

        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["get"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("test-org" in str(mock_call))

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_handles_no_organization(self, mock_project_dir: Mock, mock_organization_handler_class: Mock) -> None:
        """Test that get command handles the case when no organization is allowlisted."""
        # Setup
        mock_organization_handler, mock_handler_fns = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler
        mock_handler_fns["get_organization"].return_value = ""

        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["get"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("None" in str(mock_call))

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(
        self, mock_project_dir: Mock, mock_organization_handler_class: Mock
    ) -> None:
        """Test that get command switches directory when a project path is provided."""
        # Setup
        mock_organization_handler, _ = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["get", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.organization_handler.OrganizationHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_organization_handler_get_organization_raises(
        self, mock_project_dir: Mock, mock_organization_handler_class: Mock
    ) -> None:
        """Test that get command propagates exceptions from get_organization."""
        # Setup
        mock_organization_handler, mock_handler_fns = self.get_mock_organization_handler()
        mock_organization_handler_class.return_value = mock_organization_handler
        mock_handler_fns["get_organization"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(organization_app, ["get"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)



import unittest

from jupyter_deploy import transform_utils
from jupyter_deploy.enum import TransformType


class TestTransformFunctions(unittest.TestCase):
    """Tests for transform_utils module."""

    def test_no_transform_fn_returns_input_unchanged(self) -> None:
        """Test that _no_transform_fn returns the input value unchanged."""
        # Test with various input types
        test_cases = ["string value", 123, 1.23, True, None, ["item1", "item2"], {"key": "value"}]

        for value in test_cases:
            with self.subTest(value=value):
                result = transform_utils._no_transform_fn(value)
                self.assertEqual(result, value)
                self.assertIs(result, value, "Should return the exact same object instance")

    def test_comma_sep_str_to_list_str_handles_empty_string(self) -> None:
        """Test _comma_sep_str_to_list_str with empty input."""
        result = transform_utils._comma_sep_str_to_list_str("")
        self.assertEqual(result, [])

    def test_comma_sep_str_to_list_str_handles_single_value(self) -> None:
        """Test _comma_sep_str_to_list_str with single value."""
        result = transform_utils._comma_sep_str_to_list_str("value")
        self.assertEqual(result, ["value"])

    def test_comma_sep_str_to_list_str_handles_multiple_values(self) -> None:
        """Test _comma_sep_str_to_list_str with multiple values."""
        result = transform_utils._comma_sep_str_to_list_str("value1,value2,value3")
        self.assertEqual(result, ["value1", "value2", "value3"])

    def test_comma_sep_str_to_list_str_preserves_whitespace(self) -> None:
        """Test _comma_sep_str_to_list_str preserves whitespace."""
        result = transform_utils._comma_sep_str_to_list_str("value1, value2 , value3")
        self.assertEqual(result, ["value1", " value2 ", " value3"])

    def test_comma_sep_str_to_list_str_handles_empty_elements(self) -> None:
        """Test _comma_sep_str_to_list_str handles empty elements."""
        result = transform_utils._comma_sep_str_to_list_str("value1,,value3")
        self.assertEqual(result, ["value1", "", "value3"])

    def test_get_transform_fn_returns_correct_function(self) -> None:
        """Test get_transform_fn returns the correct function for each TransformType."""
        # Test for NO_TRANSFORM
        no_transform_fn = transform_utils.get_transform_fn(TransformType.NO_TRANSFORM)
        self.assertIs(no_transform_fn, transform_utils._no_transform_fn)

        # Test for COMMA_SEPARATED_STR_TO_LIST_STR
        comma_sep_fn = transform_utils.get_transform_fn(TransformType.COMMA_SEPARATED_STR_TO_LIST_STR)
        self.assertIs(comma_sep_fn, transform_utils._comma_sep_str_to_list_str)

    def test_all_transform_types_have_corresponding_functions(self) -> None:
        """Test that all TransformType enum values have a corresponding function in _TRANSFORM_FN_MAP."""
        for transform_type in TransformType:
            with self.subTest(transform_type=transform_type):
                self.assertIn(transform_type, transform_utils._TRANSFORM_FN_MAP)
                self.assertIsNotNone(transform_utils._TRANSFORM_FN_MAP[transform_type])
                self.assertTrue(callable(transform_utils._TRANSFORM_FN_MAP[transform_type]))



import functools
import inspect
import sys
import unittest
from collections.abc import Callable
from unittest.mock import Mock, patch

from typer.testing import CliRunner


class TestDeployCmdWithDecorator(unittest.TestCase):
    def get_mock_config_handler(self) -> tuple[Mock, dict[str, Mock]]:
        mock_config_handler = Mock()
        mock_validate = Mock()
        mock_reset_variables = Mock()
        mock_reset_secrets = Mock()
        mock_verify = Mock()
        mock_configure = Mock()
        mock_record = Mock()
        mock_has_used_preset = Mock()

        mock_config_handler.validate_and_set_preset = mock_validate
        mock_config_handler.reset_recorded_variables = mock_reset_variables
        mock_config_handler.reset_recorded_secrets = mock_reset_secrets
        mock_config_handler.verify_requirements = mock_verify
        mock_config_handler.configure = mock_configure
        mock_config_handler.record = mock_record
        mock_config_handler.has_used_preset = mock_has_used_preset

        mock_validate.return_value = True
        mock_verify.return_value = True
        mock_has_used_preset.return_value = True

        return mock_config_handler, {
            "validate_and_set_preset": mock_validate,
            "reset_recorded_variables": mock_reset_variables,
            "reset_recorded_secrets": mock_reset_secrets,
            "verify": mock_verify,
            "configure": mock_configure,
            "record": mock_record,
            "has_used_preset": mock_has_used_preset,
        }

    def get_mock_decorator(self, mock_variables: dict) -> Callable:
        def mock_decorator() -> Callable:
            def decorator(func: Callable) -> Callable:
                @functools.wraps(func)
                def wrapper(*func_args, **func_kwargs):  # type: ignore
                    # Add empty variables dict
                    func_kwargs["variables"] = mock_variables
                    return func(*func_args, **func_kwargs)

                original_sig = inspect.signature(func)
                original_params = list(original_sig.parameters.values())
                params: list[inspect.Parameter] = [p for p in original_params if p.name != "variables"]

                new_sig = original_sig.replace(parameters=params)
                wrapper.__signature__ = new_sig  # type: ignore

                return wrapper

            return decorator

        return mock_decorator

    def test_config_cmd_calls_passes_on_the_variables(self) -> None:
        if "jupyter_deploy.cli.app" in sys.modules:
            del sys.modules["jupyter_deploy.cli.app"]

        with patch("jupyter_deploy.cli.variables_decorator.with_project_variables") as mock_decorator_fn:
            mock_variables = {"variable1": Mock()}
            mock_decorator_fn.side_effect = self.get_mock_decorator(mock_variables)

            # Import AFTER patching
            from jupyter_deploy.cli.app import runner

            with patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler") as mock_config_handler:
                mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
                mock_config_handler.return_value = mock_config_handler_instance

                # Act - use the runner that has commands registered
                app_runner = CliRunner()
                result = app_runner.invoke(runner.app, ["config"])

                # Verify
                self.assertEqual(result.exit_code, 0)
                mock_decorator_fn.assert_called_once()
                mock_config_handler.assert_called_once()
                mock_config_fns["validate_and_set_preset"].assert_called_once()
                mock_config_fns["verify"].assert_called_once()
                mock_config_fns["configure"].assert_called_once_with(variable_overrides=mock_variables)
                mock_config_fns["record"].assert_called_once_with(record_vars=True, record_secrets=False)
                mock_config_fns["reset_recorded_variables"].assert_not_called()
                mock_config_fns["reset_recorded_secrets"].assert_not_called()
                mock_config_fns["has_used_preset"].assert_called_once_with("all")



import unittest
from unittest.mock import Mock, patch

from parameterized import parameterized  # type: ignore

from jupyter_deploy.engine import outdefs


class TestStrTemplateOutputDefinition(unittest.TestCase):
    @patch("jupyter_deploy.str_utils.get_trimmed_header")
    def test_str_outdef_with_value_should_get_a_valid_trimmed_header(self, mock_get_trimmed_header: Mock) -> None:
        # Arrange
        mock_get_trimmed_header.return_value = "Test description"
        outdef = outdefs.StrTemplateOutputDefinition(
            output_name="test_output", description="Test description", value="test_value"
        )

        # Act
        result = outdef.get_cli_description()

        # Assert
        mock_get_trimmed_header.assert_called_once_with("Test description")
        self.assertEqual(result, "Test description value: test_value")

    @patch("jupyter_deploy.str_utils.get_trimmed_header")
    def test_str_outdef_without_value_should_get_a_valid_trimmed_header(self, mock_get_trimmed_header: Mock) -> None:
        # Arrange
        mock_get_trimmed_header.return_value = "Test description"
        outdef = outdefs.StrTemplateOutputDefinition(
            output_name="test_output", description="Test description", value=None
        )

        # Act
        result = outdef.get_cli_description()

        # Assert
        mock_get_trimmed_header.assert_called_once_with("Test description")
        self.assertEqual(result, "Test description")


class TestRequireOutputDef(unittest.TestCase):
    def test_str_outdef_should_resolve_when_prompted_for_str(self) -> None:
        # Arrange
        output_name = "test_output"
        str_outdef = outdefs.StrTemplateOutputDefinition(
            output_name=output_name, description="Test description", value="test_value"
        )
        output_defs: dict[str, outdefs.TemplateOutputDefinition] = {output_name: str_outdef}

        # Act
        result = outdefs.require_output_def(output_defs, output_name, outdefs.StrTemplateOutputDefinition)

        # Assert
        self.assertEqual(result, str_outdef)
        self.assertEqual(result.value, "test_value")

    @parameterized.expand([(int), (float), (bool), (list), (dict)])
    def test_str_outdef_should_raise_type_exception_when_prompted_invalid_type(self, type: type) -> None:
        # Arrange
        output_name = "test_output"
        str_outdef = outdefs.StrTemplateOutputDefinition(
            output_name=output_name, description="Test description", value="test_value"
        )
        output_defs: dict[str, outdefs.TemplateOutputDefinition] = {output_name: str_outdef}

        class OtherTypeClass(outdefs.TemplateOutputDefinition[type]):  # type: ignore
            pass

        # Act & Assert
        with self.assertRaises(TypeError):
            outdefs.require_output_def(output_defs, output_name, OtherTypeClass)

    def test_should_raise_key_error_when_output_not_found(self) -> None:
        # Arrange
        output_defs: dict[str, outdefs.TemplateOutputDefinition] = {
            "existing_output": outdefs.StrTemplateOutputDefinition(output_name="existing_output")
        }
        non_existent_output = "non_existent_output"

        # Act & Assert
        with self.assertRaises(KeyError) as context:
            outdefs.require_output_def(output_defs, non_existent_output, outdefs.StrTemplateOutputDefinition)

        self.assertIn(f"Required output '{non_existent_output}' not found", str(context.exception))


class TestListStrTemplateOutputDefinition(unittest.TestCase):
    @patch("jupyter_deploy.str_utils.get_trimmed_header")
    def test_list_str_outdef_with_value_should_get_a_valid_trimmed_header(self, mock_get_trimmed_header: Mock) -> None:
        # Arrange
        mock_get_trimmed_header.return_value = "Test description"
        test_value = ["item1", "item2", "item3"]
        outdef = outdefs.ListStrTemplateOutputDefinition(
            output_name="test_output", description="Test description", value=test_value
        )

        # Act
        result = outdef.get_cli_description()

        # Assert
        mock_get_trimmed_header.assert_called_once_with("Test description")
        self.assertEqual(result, f"Test description value: {test_value}")

    @patch("jupyter_deploy.str_utils.get_trimmed_header")
    def test_list_str_outdef_with_empty_list_value(self, mock_get_trimmed_header: Mock) -> None:
        # Arrange
        mock_get_trimmed_header.return_value = "Test description"
        test_value: list[str] = []
        outdef = outdefs.ListStrTemplateOutputDefinition(
            output_name="test_output", description="Test description", value=test_value
        )

        # Act
        result = outdef.get_cli_description()

        # Assert
        mock_get_trimmed_header.assert_called_once_with("Test description")
        self.assertEqual(result, "Test description value: []")

    @patch("jupyter_deploy.str_utils.get_trimmed_header")
    def test_list_str_outdef_without_value_should_get_a_valid_trimmed_header(
        self, mock_get_trimmed_header: Mock
    ) -> None:
        # Arrange
        mock_get_trimmed_header.return_value = "Test description"
        outdef = outdefs.ListStrTemplateOutputDefinition(
            output_name="test_output", description="Test description", value=None
        )

        # Act
        result = outdef.get_cli_description()

        # Assert
        mock_get_trimmed_header.assert_called_once_with("Test description")
        self.assertEqual(result, "Test description")

    def test_list_str_outdef_should_resolve_when_prompted_for_list_str(self) -> None:
        # Arrange
        output_name = "test_output"
        test_value = ["item1", "item2", "item3"]
        list_str_outdef = outdefs.ListStrTemplateOutputDefinition(
            output_name=output_name, description="Test description", value=test_value
        )
        output_defs: dict[str, outdefs.TemplateOutputDefinition] = {output_name: list_str_outdef}

        # Act
        result = outdefs.require_output_def(output_defs, output_name, outdefs.ListStrTemplateOutputDefinition)

        # Assert
        self.assertEqual(result, list_str_outdef)
        self.assertEqual(result.value, test_value)

    def test_list_str_outdef_should_raise_type_exception_when_prompted_as_str(self) -> None:
        # Arrange
        output_name = "test_output"
        test_value = ["item1", "item2", "item3"]
        list_str_outdef = outdefs.ListStrTemplateOutputDefinition(
            output_name=output_name, description="Test description", value=test_value
        )
        output_defs: dict[str, outdefs.TemplateOutputDefinition] = {output_name: list_str_outdef}

        # Act & Assert
        with self.assertRaises(TypeError):
            outdefs.require_output_def(output_defs, output_name, outdefs.StrTemplateOutputDefinition)



import unittest
from unittest.mock import Mock, patch

from rich.console import Console

from jupyter_deploy.api.aws.ec2 import ec2_instance
from jupyter_deploy.api.aws.ec2.ec2_instance import Ec2InstanceState
from jupyter_deploy.provider.aws.aws_ec2_runner import AwsEc2Instruction, AwsEc2Runner
from jupyter_deploy.provider.instruction_runner import InterruptInstructionError
from jupyter_deploy.provider.resolved_argdefs import ResolvedInstructionArgument, StrResolvedInstructionArgument


class TestAwsEc2Runner(unittest.TestCase):
    @patch("boto3.client")
    def test_aws_ec2_runner_instantiates_client(self, mock_boto3_client: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_boto3_client.return_value = mock_client

        # Execute
        runner = AwsEc2Runner(region_name="us-west-2")

        # Assert
        mock_boto3_client.assert_called_once_with("ec2", region_name="us-west-2")
        self.assertEqual(runner.client, mock_client)

    def test_aws_ec2_raise_not_implemented_error_on_unmatched_instruction_name(self) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Execute & Assert
        with self.assertRaises(NotImplementedError) as context:
            runner.execute_instruction(
                instruction_name="non-existent-instruction", resolved_arguments={}, console=console
            )

        self.assertIn("non-existent-instruction", str(context.exception))


class TestDescribeInstanceStatus(unittest.TestCase):
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    def test_happy_path(self, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "running", "Code": 16}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute
        result = runner._describe_instance_status(resolved_arguments=resolved_args, console=console)

        # Assert
        mock_describe_instance_status.assert_called_once_with(runner.client, instance_id="i-123456789abcdef")
        console.print.assert_called()
        self.assertEqual(result["InstanceStateName"].value, "running")

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    def test_raises_when_describe_instance_status_raises(self, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)
        mock_describe_instance_status.side_effect = ValueError("Instance not found")

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(ValueError) as context:
            runner._describe_instance_status(resolved_arguments=resolved_args, console=console)

        self.assertIn("Instance not found", str(context.exception))
        mock_describe_instance_status.assert_called_once()


class TestStartInstance(unittest.TestCase):
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.start_instance")
    def test_happy_path_on_stopped_state(self, mock_start_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "stopped", "Code": 80}}
        mock_start_instance.return_value = {
            "CurrentState": {"Name": "pending", "Code": 0},
            "PreviousState": {"Name": "stopped", "Code": 80},
        }

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute
        result = runner._start_instance(resolved_arguments=resolved_args, console=console)

        # Assert
        mock_describe_instance_status.assert_called_once_with(runner.client, instance_id="i-123456789abcdef")
        mock_start_instance.assert_called_once_with(runner.client, instance_id="i-123456789abcdef")
        console.print.assert_called()
        self.assertEqual(result, {})

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.start_instance")
    def test_interrupt_on_pending_state(self, mock_start_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "pending", "Code": 0}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._start_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_start_instance.assert_not_called()
        console.print.assert_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.start_instance")
    def test_interrupt_on_running_state(self, mock_start_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "running", "Code": 16}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._start_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_start_instance.assert_not_called()
        console.print.assert_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.start_instance")
    def test_interrupt_on_stopping_state(self, mock_start_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "stopping", "Code": 64}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._start_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_start_instance.assert_not_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.start_instance")
    def test_interrupt_on_shutting_down_state(
        self, mock_start_instance: Mock, mock_describe_instance_status: Mock
    ) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "shutting-down", "Code": 32}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._start_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_start_instance.assert_not_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.start_instance")
    def test_interrupt_on_terminated_state(
        self, mock_start_instance: Mock, mock_describe_instance_status: Mock
    ) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "terminated", "Code": 48}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._start_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_start_instance.assert_not_called()


class TestStopInstance(unittest.TestCase):
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.stop_instance")
    def test_happy_path_on_running_state(self, mock_stop_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "running", "Code": 16}}
        mock_stop_instance.return_value = {
            "CurrentState": {"Name": "stopping", "Code": 64},
            "PreviousState": {"Name": "running", "Code": 16},
        }

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute
        result = runner._stop_instance(resolved_arguments=resolved_args, console=console)

        # Assert
        mock_describe_instance_status.assert_called_once_with(runner.client, instance_id="i-123456789abcdef")
        mock_stop_instance.assert_called_once_with(runner.client, instance_id="i-123456789abcdef")
        console.print.assert_called()
        self.assertEqual(result, {})

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.stop_instance")
    def test_interrupt_on_pending_state(self, mock_stop_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "pending", "Code": 0}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._stop_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_stop_instance.assert_not_called()
        console.print.assert_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.stop_instance")
    def test_interrupt_on_stopping_state(self, mock_stop_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "stopping", "Code": 64}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._stop_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_stop_instance.assert_not_called()
        console.print.assert_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.stop_instance")
    def test_interrupt_on_stopped_state(self, mock_stop_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "stopped", "Code": 80}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._stop_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_stop_instance.assert_not_called()
        console.print.assert_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.stop_instance")
    def test_interrupt_on_shutting_down_state(
        self, mock_stop_instance: Mock, mock_describe_instance_status: Mock
    ) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "shutting-down", "Code": 32}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._stop_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_stop_instance.assert_not_called()
        console.print.assert_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.stop_instance")
    def test_interrupt_on_terminated_state(self, mock_stop_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "terminated", "Code": 48}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._stop_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_stop_instance.assert_not_called()
        console.print.assert_called()


class TestRebootInstance(unittest.TestCase):
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.restart_instance")
    def test_happy_path_on_running_state(
        self, mock_restart_instance: Mock, mock_describe_instance_status: Mock
    ) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "running", "Code": 16}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute
        result = runner._reboot_instance(resolved_arguments=resolved_args, console=console)

        # Assert
        mock_describe_instance_status.assert_called_once_with(runner.client, instance_id="i-123456789abcdef")
        mock_restart_instance.assert_called_once_with(runner.client, instance_id="i-123456789abcdef")
        console.print.assert_called()
        self.assertEqual(result, {})

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.restart_instance")
    def test_interrupt_on_pending_state(self, mock_restart_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "pending", "Code": 0}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._reboot_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_restart_instance.assert_not_called()
        console.print.assert_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.restart_instance")
    def test_interrupt_on_stopping_state(
        self, mock_restart_instance: Mock, mock_describe_instance_status: Mock
    ) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "stopping", "Code": 64}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._reboot_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_restart_instance.assert_not_called()
        console.print.assert_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.restart_instance")
    def test_interrupt_on_stopped_state(self, mock_restart_instance: Mock, mock_describe_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "stopped", "Code": 80}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._reboot_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_restart_instance.assert_not_called()
        console.print.assert_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.restart_instance")
    def test_interrupt_on_shutting_down_state(
        self, mock_restart_instance: Mock, mock_describe_instance_status: Mock
    ) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "shutting-down", "Code": 32}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._reboot_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_restart_instance.assert_not_called()
        console.print.assert_called()

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.describe_instance_status")
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.restart_instance")
    def test_interrupt_on_terminated_state(
        self, mock_restart_instance: Mock, mock_describe_instance_status: Mock
    ) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mocks
        mock_describe_instance_status.return_value = {"InstanceState": {"Name": "terminated", "Code": 48}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(InterruptInstructionError):
            runner._reboot_instance(resolved_arguments=resolved_args, console=console)

        mock_describe_instance_status.assert_called_once()
        mock_restart_instance.assert_not_called()
        console.print.assert_called()


class TestWaitForState(unittest.TestCase):
    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.poll_for_instance_status")
    def test_happy_path(self, mock_poll_for_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mock
        mock_poll_for_instance_status.return_value = {"InstanceState": {"Name": "running", "Code": 16}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute
        result = runner._wait_for_state(
            resolved_arguments=resolved_args, console=console, desired_state=Ec2InstanceState.RUNNING
        )

        # Assert
        mock_poll_for_instance_status.assert_called_once_with(
            runner.client,
            console=console,
            instance_id="i-123456789abcdef",
            desired_state=Ec2InstanceState.RUNNING,
            timeout_seconds=60,  # default timeout
        )
        self.assertEqual(result["InstanceStateName"].value, "running")

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.poll_for_instance_status")
    def test_allows_timeout_override(self, mock_poll_for_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mock
        mock_poll_for_instance_status.return_value = {"InstanceState": {"Name": "stopped", "Code": 80}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute
        result = runner._wait_for_state(
            resolved_arguments=resolved_args,
            console=console,
            desired_state=Ec2InstanceState.STOPPED,
            timeout_seconds=120,  # custom timeout
        )

        # Assert
        mock_poll_for_instance_status.assert_called_once_with(
            runner.client,
            console=console,
            instance_id="i-123456789abcdef",
            desired_state=Ec2InstanceState.STOPPED,
            timeout_seconds=120,  # custom timeout
        )
        self.assertEqual(result["InstanceStateName"].value, "stopped")

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.poll_for_instance_status")
    def test_raises_on_poll_raises(self, mock_poll_for_instance_status: Mock) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mock to raise an exception
        mock_poll_for_instance_status.side_effect = TimeoutError("Timed out waiting for instance")

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute & Assert
        with self.assertRaises(TimeoutError) as context:
            runner._wait_for_state(
                resolved_arguments=resolved_args, console=console, desired_state=Ec2InstanceState.RUNNING
            )

        self.assertIn("Timed out waiting for instance", str(context.exception))
        mock_poll_for_instance_status.assert_called_once()


class TestExecuteInstructions(unittest.TestCase):
    def test_all_instructions_implemented(self) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)
        resolved_args: dict[str, ResolvedInstructionArgument] = {}

        # Create patch targets for all methods that would be called by execute_instruction
        patches = [
            patch.object(runner, "_describe_instance_status", return_value={}),
            patch.object(runner, "_start_instance", return_value={}),
            patch.object(runner, "_stop_instance", return_value={}),
            patch.object(runner, "_reboot_instance", return_value={}),
            patch.object(runner, "_wait_for_state", return_value={}),
        ]

        instruction_method_map = {
            AwsEc2Instruction.DESCRIBE_INSTANCE_STATUS: "_describe_instance_status",
            AwsEc2Instruction.START_INSTANCE: "_start_instance",
            AwsEc2Instruction.STOP_INSTANCE: "_stop_instance",
            AwsEc2Instruction.REBOOT_INSTANCE: "_reboot_instance",
            AwsEc2Instruction.WAIT_FOR_RUNNING: "_wait_for_state",
            AwsEc2Instruction.WAIT_FOR_STOPPED: "_wait_for_state",
        }

        # Test each instruction
        for instruction, method_name in instruction_method_map.items():
            with self.subTest(instruction=instruction):
                # Start all patches
                mocks = [p.start() for p in patches]
                try:
                    # Execute
                    runner.execute_instruction(
                        instruction_name=instruction, resolved_arguments=resolved_args, console=console
                    )

                    # Assert the correct method was called
                    expected_mock = next(m for m in mocks if m._mock_name == method_name)
                    expected_mock.assert_called_once()

                    # Assert other methods were not called
                    for mock in mocks:
                        if mock._mock_name != method_name:
                            mock.assert_not_called()

                finally:
                    # Stop all patches
                    for p in patches:
                        p.stop()

    def test_raise_not_implemented_error_on_unrecognized_instruction(self) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)
        resolved_args: dict[str, ResolvedInstructionArgument] = {}

        # Execute & Assert
        with self.assertRaises(NotImplementedError) as context:
            runner.execute_instruction(
                instruction_name="unknown-instruction", resolved_arguments=resolved_args, console=console
            )

        self.assertIn("unknown-instruction", str(context.exception))

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.poll_for_instance_status")
    def test_wait_for_running_pass_a_timeout_of_at_least_sixty_seconds(
        self, mock_poll_for_instance_status: Mock
    ) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mock
        mock_poll_for_instance_status.return_value = {"InstanceState": {"Name": "running", "Code": 16}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute
        runner.execute_instruction(
            instruction_name=AwsEc2Instruction.WAIT_FOR_RUNNING, resolved_arguments=resolved_args, console=console
        )

        # Assert that poll_for_instance_status was called with a timeout >= 60 seconds
        mock_poll_for_instance_status.assert_called_once()
        actual_timeout = mock_poll_for_instance_status.call_args[1]["timeout_seconds"]
        self.assertGreaterEqual(
            actual_timeout,
            60,
            f"WAIT_FOR_RUNNING should use a timeout of at least 60 seconds, but got {actual_timeout}",
        )

        # Verify we're using the correct state
        self.assertEqual(
            mock_poll_for_instance_status.call_args[1]["desired_state"], ec2_instance.Ec2InstanceState.RUNNING
        )

    @patch("jupyter_deploy.api.aws.ec2.ec2_instance.poll_for_instance_status")
    def test_wait_for_stopped_pass_a_timeout_of_at_least_five_minutes(
        self, mock_poll_for_instance_status: Mock
    ) -> None:
        # Setup
        runner = AwsEc2Runner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Configure mock
        mock_poll_for_instance_status.return_value = {"InstanceState": {"Name": "stopped", "Code": 80}}

        # Prepare arguments
        instance_id_arg = StrResolvedInstructionArgument(argument_name="instance_id", value="i-123456789abcdef")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"instance_id": instance_id_arg}

        # Execute
        runner.execute_instruction(
            instruction_name=AwsEc2Instruction.WAIT_FOR_STOPPED, resolved_arguments=resolved_args, console=console
        )

        # Assert that poll_for_instance_status was called with a timeout >= 300 seconds (5 minutes)
        mock_poll_for_instance_status.assert_called_once()
        actual_timeout = mock_poll_for_instance_status.call_args[1]["timeout_seconds"]
        self.assertGreaterEqual(
            actual_timeout,
            300,
            f"WAIT_FOR_STOPPED should use a timeout of at least 5 minutes (300 seconds), but got {actual_timeout}",
        )

        # Verify we're using the correct state
        self.assertEqual(
            mock_poll_for_instance_status.call_args[1]["desired_state"], ec2_instance.Ec2InstanceState.STOPPED
        )



import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from jupyter_deploy.engine.terraform.tf_variables import TerraformVariablesHandler


class TestTerraformVariablesHandler(unittest.TestCase):
    def test_successfully_instantiates(self) -> None:
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)
        self.assertEqual(handler.project_path, project_path)
        self.assertEqual(handler.project_manifest, manifest)

    def test_get_recorded_variables_filepath(self) -> None:
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Get the filepath and verify it's correct
        result = handler.get_recorded_variables_filepath()

        # Should be project_path/engine/jdinputs.auto.tfvars
        self.assertEqual(result, project_path / "engine" / "jdinputs.auto.tfvars")

    def test_get_recorded_secrets_filepath(self) -> None:
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Get the filepath and verify it's correct
        result = handler.get_recorded_secrets_filepath()

        # Should be project_path/engine/jdinputs.secrets.auto.tfvars
        self.assertEqual(result, project_path / "engine" / "jdinputs.secrets.auto.tfvars")


class TestIsTemplateDir(unittest.TestCase):
    @patch("jupyter_deploy.fs_utils.file_exists")
    def test_return_true_when_variables_dot_tf_exists(self, mock_file_exists: Mock) -> None:
        mock_file_exists.return_value = True
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)

        result = handler.is_template_directory()
        self.assertTrue(result)
        mock_file_exists.assert_called_once_with(project_path / "engine" / "variables.tf")

    @patch("jupyter_deploy.fs_utils.file_exists")
    def test_return_false_when_variables_dot_tf_does_not_exists(self, mock_file_exists: Mock) -> None:
        mock_file_exists.return_value = False
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)

        result = handler.is_template_directory()
        self.assertFalse(result)
        mock_file_exists.assert_called_once_with(project_path / "engine" / "variables.tf")


class TestGetTemplateVariables(unittest.TestCase):
    @patch("jupyter_deploy.fs_utils.read_short_file")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_variables_dot_tf_content")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_dot_tfvars_content_and_add_defaults")
    def test_returns_variables(
        self, mock_parse_tfvars: Mock, mock_parse_variables: Mock, mock_read_short_file: Mock
    ) -> None:
        project_path = Path("/mock/project")
        manifest = Mock()

        mock_read_short_file.side_effect = ["content-1", "content-2"]

        # mock terraform vars instances and their to_template_definition() method
        mock_1 = Mock()
        mock_2 = Mock()
        mock_3 = Mock()
        mock_to_template_def_1 = Mock()
        mock_to_template_def_2 = Mock()
        mock_to_template_def_3 = Mock()
        mock_1.to_template_definition = mock_to_template_def_1
        mock_2.to_template_definition = mock_to_template_def_2
        mock_3.to_template_definition = mock_to_template_def_3
        mock_to_template_def_1.return_value = {"val": "1"}
        mock_to_template_def_2.return_value = {"val": "2"}
        mock_to_template_def_3.return_value = {"val": "3"}

        # mock the variable parsing response
        mock_vars_from_vars_dot_tf = {"var1": mock_1, "var2": mock_2}
        mock_parse_variables.return_value = mock_vars_from_vars_dot_tf

        # make the tfvars modify the 2nd key value only
        def tfvars_side_effect(*largs, **kwargs) -> None:  # type: ignore
            input_vars = kwargs["variable_defs"]
            assert type(input_vars) is dict
            input_vars.update({"var2": mock_3})

        mock_parse_tfvars.side_effect = tfvars_side_effect

        # Act
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)
        result = handler.get_template_variables()

        # Assert
        self.assertEqual(result, {"var1": {"val": "1"}, "var2": {"val": "3"}})

        # should have read both vars.tf and .tfvars files
        self.assertEqual(mock_read_short_file.call_count, 2)
        self.assertEqual(mock_read_short_file.mock_calls[0][1][0], project_path / "engine" / "variables.tf")
        self.assertEqual(
            mock_read_short_file.mock_calls[1][1][0], project_path / "engine" / "presets" / "defaults-all.tfvars"
        )

        # should have parsed with the appropriate content
        mock_parse_variables.assert_called_once_with("content-1")
        mock_parse_tfvars.assert_called_once_with("content-2", variable_defs=mock_vars_from_vars_dot_tf)

        # only the final variable wrappers should have called their convert method
        mock_to_template_def_1.assert_called_once()
        mock_to_template_def_2.assert_not_called()
        mock_to_template_def_3.assert_called_once()

    @patch("jupyter_deploy.fs_utils.read_short_file")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_variables_dot_tf_content")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_dot_tfvars_content_and_add_defaults")
    def test_raises_on_large_variables_dot_tf_file(
        self, mock_parse_tfvars: Mock, mock_parse_variables: Mock, mock_read_short_file: Mock
    ) -> None:
        # Prepare
        mock_read_short_file.side_effect = RuntimeError("File is too large!")

        # Act
        handler = TerraformVariablesHandler(project_path=Path("/mock/project"), project_manifest=Mock())
        with self.assertRaises(RuntimeError):
            handler.get_template_variables()

        # Verify
        mock_read_short_file.assert_called_once()
        mock_parse_tfvars.assert_not_called()
        mock_parse_variables.assert_not_called()

    @patch("jupyter_deploy.fs_utils.read_short_file")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_variables_dot_tf_content")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_dot_tfvars_content_and_add_defaults")
    def test_raises_on_variables_dot_tf_read_error(
        self, mock_parse_tfvars: Mock, mock_parse_variables: Mock, mock_read_short_file: Mock
    ) -> None:
        # Prepare
        mock_read_short_file.side_effect = ["content-1", RuntimeError("File is too large!")]
        mock_parse_variables.return_value = {}

        # Act
        handler = TerraformVariablesHandler(project_path=Path("/mock/project"), project_manifest=Mock())
        with self.assertRaises(RuntimeError):
            handler.get_template_variables()

        # Verify
        mock_read_short_file.assert_called()
        mock_parse_variables.assert_called_once()
        mock_parse_tfvars.assert_not_called()

    @patch("jupyter_deploy.fs_utils.read_short_file")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_variables_dot_tf_content")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_dot_tfvars_content_and_add_defaults")
    def test_raises_tfvars_read_error(
        self, mock_parse_tfvars: Mock, mock_parse_variables: Mock, mock_read_short_file: Mock
    ) -> None:
        # Prepare
        mock_read_short_file.side_effect = ["content-1", "content-2"]
        mock_1 = Mock()
        mock_to_template_def_1 = Mock()
        mock_1.to_template_definition = mock_to_template_def_1
        mock_parse_variables.return_value = {"val1": mock_1}
        mock_to_template_def_1.return_value = {"val": "1"}

        # Act
        handler = TerraformVariablesHandler(project_path=Path("/mock/project"), project_manifest=Mock())
        result1 = handler.get_template_variables()

        # Verify-1
        self.assertEqual(result1, {"val1": {"val": "1"}})
        self.assertEqual(mock_read_short_file.call_count, 2)
        self.assertEqual(mock_parse_variables.call_count, 1)
        self.assertEqual(mock_parse_tfvars.call_count, 1)
        self.assertEqual(mock_to_template_def_1.call_count, 1)

        # Act again
        result2 = handler.get_template_variables()
        self.assertEqual(result1, result2)
        self.assertEqual(mock_read_short_file.call_count, 2)
        self.assertEqual(mock_parse_variables.call_count, 1)
        self.assertEqual(mock_parse_tfvars.call_count, 1)
        self.assertEqual(mock_to_template_def_1.call_count, 1)


class TestUpdateVariablesRecord(unittest.TestCase):
    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("jupyter_deploy.fs_utils.read_short_file")
    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_and_update_dot_tfvars_content")
    def test_happy_case_when_tfvars_exists(
        self, mock_get_updated_vars: Mock, mock_write_file: Mock, mock_read_file: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        mock_file_exists.return_value = True
        mock_read_file.return_value = "existing_content"
        mock_get_updated_vars.return_value = ["updated_line1", "updated_line2"]

        # Create a handler with mocked template variables
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=Mock())
        mock_validate1 = Mock()
        mock_validate2 = Mock()
        handler._template_vars = {
            "var1": Mock(validate_value=mock_validate1),
            "var2": Mock(validate_value=mock_validate2),
        }

        # Execute
        handler.update_variable_records({"var1": "value1", "var2": "value2"})

        # Assert
        mock_file_exists.assert_called_once_with(project_path / "engine" / "jdinputs.auto.tfvars")
        mock_read_file.assert_called_once_with(project_path / "engine" / "jdinputs.auto.tfvars")

        mock_get_updated_vars.assert_called_once()
        mock_write_file.assert_called_once_with(
            project_path / "engine" / "jdinputs.auto.tfvars", ["updated_line1", "updated_line2"]
        )

        mock_validate1.assert_called_once_with("value1")
        mock_validate2.assert_called_once_with("value2")

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("jupyter_deploy.fs_utils.read_short_file")
    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_and_update_dot_tfvars_content")
    def test_happy_case_when_tfvars_does_not_exist(
        self, mock_get_updated_vars: Mock, mock_write_file: Mock, mock_read_file: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        mock_file_exists.return_value = False
        mock_get_updated_vars.return_value = ["line1", "line2"]

        # Create a handler with mocked template variables
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=Mock())
        mock_validate = Mock()
        handler._template_vars = {
            "var1": Mock(validate_value=mock_validate),
        }

        # Execute
        handler.update_variable_records({"var1": "new_value"})

        # Assert
        mock_file_exists.assert_called_once_with(project_path / "engine" / "jdinputs.auto.tfvars")

        mock_get_updated_vars.assert_called_once()
        mock_write_file.assert_called_once_with(project_path / "engine" / "jdinputs.auto.tfvars", ["line1", "line2"])
        mock_validate.assert_called_once_with("new_value")
        mock_read_file.assert_not_called()

    def test_raises_without_update_if_any_variable_is_not_of_the_right_type(self) -> None:
        # Setup
        project_path = Path("/mock/project")

        # Create a handler with mocked template variables
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=Mock())

        # Mock that will raise TypeError
        mock_var = Mock()
        mock_var.validate_value.side_effect = TypeError("Invalid type")

        mock_validate1 = Mock()
        handler._template_vars = {
            "var1": mock_var,
            "var2": Mock(validate_value=mock_validate1),
        }

        # Execute & Assert
        with self.assertRaises(TypeError):
            handler.update_variable_records({"var1": "invalid_value", "var2": "valid_value"})

        # Verify validate_value was called but we didn't proceed to process other vars
        mock_var.validate_value.assert_called_once_with("invalid_value")
        mock_validate1.assert_not_called()

    def test_raises_without_update_if_any_variable_is_not_found(self) -> None:
        # Setup
        project_path = Path("/mock/project")

        # Create a handler with mocked template variables
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=Mock())
        handler._template_vars = {
            "var1": Mock(validate_value=Mock()),
        }

        # Execute & Assert
        with self.assertRaises(KeyError):
            handler.update_variable_records({"var1": "value1", "nonexistent_var": "value2"})

        # No assertions on validate_value needed as the key error would happen before validation

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("jupyter_deploy.fs_utils.read_short_file")
    def test_raises_if_reading_tfvars_file_fails(self, mock_read_file: Mock, mock_file_exists: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        mock_file_exists.return_value = True
        mock_read_file.side_effect = RuntimeError("File read error")

        # Create a handler with mocked template variables
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=Mock())
        handler._template_vars = {
            "var1": Mock(validate_value=Mock()),
        }

        # Execute & Assert
        with self.assertRaises(RuntimeError):
            handler.update_variable_records({"var1": "value1"})

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("jupyter_deploy.fs_utils.read_short_file")
    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_and_update_dot_tfvars_content")
    def test_raises_if_write_file_raises(
        self, mock_get_updated_vars: Mock, mock_write_file: Mock, mock_read_file: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        mock_file_exists.return_value = True
        mock_read_file.return_value = "existing_content"
        mock_get_updated_vars.return_value = ["line1", "line2"]
        mock_write_file.side_effect = RuntimeError("File write error")

        # Create a handler with mocked template variables
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=Mock())
        handler._template_vars = {
            "var1": Mock(validate_value=Mock()),
        }

        # Execute & Assert
        with self.assertRaises(RuntimeError):
            handler.update_variable_records({"var1": "value1"})

    def test_noop_when_passed_empty_dict(self) -> None:
        # Setup
        project_path = Path("/mock/project")

        # Create a handler with mocked get_template_variables
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=Mock())
        handler.get_template_variables = Mock()  # type: ignore

        # Execute
        handler.update_variable_records({})

        # Assert - get_template_variables should not be called
        handler.get_template_variables.assert_not_called()


class TestCreateFilteredPresetFile(unittest.TestCase):
    @patch("jupyter_deploy.fs_utils.read_short_file")
    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_and_remove_overridden_variables_from_content")
    def test_read_preset_retrieve_variables_filter_write_and_return_new_path(
        self, mock_parse_and_remove: Mock, mock_write_file: Mock, mock_read_file: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock the retrieved variable names
        handler.get_variable_names_assigned_in_config = Mock(return_value=["var1", "var2"])  # type: ignore

        # Mock the file read and parsing
        mock_read_file.return_value = "preset_content"
        mock_parse_and_remove.return_value = ["filtered_line1", "filtered_line2"]

        # Execute
        base_preset_path = Path("/mock/project/presets/example.tfvars")
        result = handler.create_filtered_preset_file(base_preset_path)

        # Assert
        mock_read_file.assert_called_once_with(base_preset_path)
        mock_parse_and_remove.assert_called_once_with("preset_content", ["var1", "var2"])
        mock_write_file.assert_called_once_with(
            project_path / "engine" / "jdinputs.preset.override.tfvars", ["filtered_line1", "filtered_line2"]
        )
        self.assertEqual(result, project_path / "engine" / "jdinputs.preset.override.tfvars")

    @patch("jupyter_deploy.fs_utils.read_short_file")
    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_and_remove_overridden_variables_from_content")
    def test_does_not_write_when_all_content_gets_filtered_out(
        self, mock_parse_and_remove: Mock, mock_write_file: Mock, mock_read_file: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock the retrieved variable names
        handler.get_variable_names_assigned_in_config = Mock(return_value=["var1", "var2"])  # type: ignore

        # Mock the file read and parsing to return empty list (all filtered out)
        mock_read_file.return_value = "preset_content"
        mock_parse_and_remove.return_value = []

        # Execute
        base_preset_path = Path("/mock/project/presets/example.tfvars")
        result = handler.create_filtered_preset_file(base_preset_path)

        # Assert
        mock_read_file.assert_called_once_with(base_preset_path)
        mock_parse_and_remove.assert_called_once_with("preset_content", ["var1", "var2"])
        mock_write_file.assert_not_called()  # This is the key assertion - file shouldn't be written
        self.assertEqual(result, project_path / "engine" / "jdinputs.preset.override.tfvars")

    @patch("jupyter_deploy.fs_utils.read_short_file")
    def test_raises_when_preset_does_not_exist(self, mock_read_file: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock file read to raise FileNotFoundError
        mock_read_file.side_effect = FileNotFoundError("File not found")

        # Execute & Assert
        base_preset_path = Path("/mock/project/presets/nonexistent.tfvars")
        with self.assertRaises(FileNotFoundError):
            handler.create_filtered_preset_file(base_preset_path)

    @patch("jupyter_deploy.fs_utils.read_short_file")
    def test_raises_when_preset_read_raises(self, mock_read_file: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock file read to raise some other error
        mock_read_file.side_effect = RuntimeError("Error reading file")

        # Execute & Assert
        base_preset_path = Path("/mock/project/presets/example.tfvars")
        with self.assertRaises(RuntimeError):
            handler.create_filtered_preset_file(base_preset_path)

    @patch("jupyter_deploy.fs_utils.read_short_file")
    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_varfiles.parse_and_remove_overridden_variables_from_content")
    def test_raises_when_file_write_raises(
        self, mock_parse_and_remove: Mock, mock_write_file: Mock, mock_read_file: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock the retrieved variable names
        handler.get_variable_names_assigned_in_config = Mock(return_value=["var1"])  # type: ignore

        # Mock the file read and parsing
        mock_read_file.return_value = "preset_content"
        mock_parse_and_remove.return_value = ["filtered_line"]

        # Mock write to raise an error
        mock_write_file.side_effect = OSError("Permission denied")

        # Execute & Assert
        base_preset_path = Path("/mock/project/presets/example.tfvars")
        with self.assertRaises(OSError):
            handler.create_filtered_preset_file(base_preset_path)


class TestResetRecordedVariables(unittest.TestCase):
    @patch("jupyter_deploy.engine.engine_variables.EngineVariablesHandler.reset_recorded_variables")
    @patch("jupyter_deploy.fs_utils.delete_file_if_exists")
    def test_calls_parent_method_and_delete(self, mock_delete_file: Mock, mock_parent_reset: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)
        mock_delete_file.return_value = False  # File wasn't deleted

        # Execute
        handler.reset_recorded_variables()

        # Assert parent method was called and delete_file was called with correct path
        mock_parent_reset.assert_called_once()
        mock_delete_file.assert_called_once_with(project_path / "engine" / "jdinputs.auto.tfvars")

    @patch("rich.console.Console")
    @patch("jupyter_deploy.engine.engine_variables.EngineVariablesHandler.reset_recorded_variables")
    @patch("jupyter_deploy.fs_utils.delete_file_if_exists")
    def test_calls_console_and_print_on_actual_deletion(
        self, mock_delete_file: Mock, mock_parent_reset: Mock, mock_console_cls: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)
        mock_delete_file.return_value = True  # File was deleted
        mock_console = Mock()
        mock_console_cls.return_value = mock_console

        # Execute
        handler.reset_recorded_variables()

        # Assert
        mock_parent_reset.assert_called_once()
        mock_delete_file.assert_called_once()
        mock_console.print.assert_called_once()
        # Check that the console message contains the filename
        self.assertIn("jdinputs.auto.tfvars", mock_console.print.call_args[0][0])

    @patch("jupyter_deploy.engine.engine_variables.EngineVariablesHandler.reset_recorded_variables")
    @patch("jupyter_deploy.fs_utils.delete_file_if_exists")
    def test_raises_os_error_when_delete_file_raises_os_error(
        self, mock_delete_file: Mock, mock_parent_reset: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)
        mock_delete_file.side_effect = OSError("Permission denied")

        # Execute & Assert
        with self.assertRaises(OSError):
            handler.reset_recorded_variables()

        mock_parent_reset.assert_called_once()
        mock_delete_file.assert_called_once()


class TestResetRecordedSecrets(unittest.TestCase):
    @patch("jupyter_deploy.engine.engine_variables.EngineVariablesHandler.reset_recorded_secrets")
    @patch("jupyter_deploy.fs_utils.delete_file_if_exists")
    def test_calls_parent_method_and_delete(self, mock_delete_file: Mock, mock_parent_reset: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)
        mock_delete_file.return_value = False  # File wasn't deleted

        # Execute
        handler.reset_recorded_secrets()

        # Assert parent method was called and delete_file was called with correct path
        mock_parent_reset.assert_called_once()
        mock_delete_file.assert_called_once_with(project_path / "engine" / "jdinputs.secrets.auto.tfvars")

    @patch("rich.console.Console")
    @patch("jupyter_deploy.engine.engine_variables.EngineVariablesHandler.reset_recorded_secrets")
    @patch("jupyter_deploy.fs_utils.delete_file_if_exists")
    def test_calls_console_and_print_on_actual_deletion(
        self, mock_delete_file: Mock, mock_parent_reset: Mock, mock_console_cls: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)
        mock_delete_file.return_value = True  # File was deleted
        mock_console = Mock()
        mock_console_cls.return_value = mock_console

        # Execute
        handler.reset_recorded_secrets()

        # Assert
        mock_parent_reset.assert_called_once()
        mock_delete_file.assert_called_once()
        mock_console.print.assert_called_once()
        # Check that the console message contains the filename
        self.assertIn("jdinputs.secrets.auto.tfvars", mock_console.print.call_args[0][0])

    @patch("jupyter_deploy.engine.engine_variables.EngineVariablesHandler.reset_recorded_secrets")
    @patch("jupyter_deploy.fs_utils.delete_file_if_exists")
    def test_raises_os_error_when_delete_file_raises_os_error(
        self, mock_delete_file: Mock, mock_parent_reset: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = TerraformVariablesHandler(project_path=project_path, project_manifest=manifest)
        mock_delete_file.side_effect = OSError("Permission denied")

        # Execute & Assert
        with self.assertRaises(OSError):
            handler.reset_recorded_secrets()

        mock_parent_reset.assert_called_once()
        mock_delete_file.assert_called_once()



from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict

S = TypeVar("S")


class ResolvedInstructionResult(BaseModel, Generic[S]):
    """Wrapper class for a value returned by a successful instruction."""

    model_config = ConfigDict(extra="allow")
    result_name: str
    value: S


class StrResolvedInstructionResult(ResolvedInstructionResult[str]):
    """Wrapper class for string value returned by a successful instruction."""

    pass


class IntResolvedInstructionResult(ResolvedInstructionResult[int]):
    """Wrapper class for int value returned by a successful instruction."""

    pass


class ListStrResolvedInstructionResult(ResolvedInstructionResult[list[str]]):
    """Wrapper class for list of strings value returned by a successful instruction."""

    pass



from pathlib import Path

from rich import console as rich_console

from jupyter_deploy.engine import outdefs
from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.manifest import JupyterDeployManifest


class EngineOpenHandler:
    """Base class for engine-specific open handlers."""

    def __init__(
        self,
        project_path: Path,
        project_manifest: JupyterDeployManifest,
        output_handler: EngineOutputsHandler,
    ) -> None:
        """Instantiate the base open handler."""
        self.project_path = project_path
        self.project_manifest = project_manifest
        self.output_handler = output_handler
        self._console: rich_console.Console | None = None

    def get_console(self) -> rich_console.Console:
        """Return the instance's rich console."""
        if self._console:
            return self._console
        self._console = rich_console.Console()
        return self._console

    def get_url(self) -> str:
        """Return the URL to access the notebook app, or the empty string if it cannot be resolved."""
        try:
            url_outdef = self.output_handler.get_declared_output_def("open_url", outdefs.StrTemplateOutputDefinition)
        except (ValueError, NotImplementedError, TypeError) as e:
            console = self.get_console()
            console.print(
                f":x: Could not retrieve the declared value 'open_url' in the manifest. Error details: {e}",
                style="red",
            )
            return ""
        except KeyError as _:
            console = self.get_console()
            console.print(":warning: URL not available.", style="yellow")
            console.print("This is normal if you have not deployed the project.", style="yellow")
            console.line()
            console.print("Run [bold cyan]jd config[/] then [bold cyan]jd up[/].")
            return ""

        if not url_outdef.value:
            console = self.get_console()
            console.print(
                ":x: Could not get the resolved output value for 'open_url'. "
                "Have you run `jd up` from the project directory?",
                style="red",
            )
            return ""

        return url_outdef.value



from pathlib import Path

from rich.console import Console

from jupyter_deploy.engine.engine_up import EngineUpHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_up
from jupyter_deploy.handlers.base_project_handler import BaseProjectHandler


class UpHandler(BaseProjectHandler):
    _handler: EngineUpHandler

    def __init__(self) -> None:
        """Base class to manage the up command of a jupyter-deploy project."""
        super().__init__()

        if self.engine == EngineType.TERRAFORM:
            self._handler = tf_up.TerraformUpHandler(project_path=self.project_path)
        else:
            raise NotImplementedError(f"UpHandler implementation not found for engine: {self.engine}")

    def get_default_config_filename(self) -> str:
        """Get the default config file name for the current engine."""
        return self._handler.get_default_config_filename()

    def get_config_file_path(self, config_filename: str | None = None) -> Path:
        """Get the full path to the config file."""
        console = Console()

        if config_filename is None:
            config_filename = self.get_default_config_filename()

        config_file_path = self._handler.engine_dir_path / config_filename

        if not config_file_path.exists():
            console.print(
                f":x: Config file '{config_filename}' not found in {self.project_path}. "
                f"If you have not yet generated a config file for your current project, "
                f'please run "jd config" from the project directory first.',
                style="red",
            )
            return Path.cwd()

        return config_file_path

    def apply(self, config_file_path: Path, auto_approve: bool = False) -> None:
        """Apply the infrastructure changes defined in the config file.

        Args:
            config_file_path: The path to the config file.
            auto_approve: Whether to auto-approve the changes without prompting.
        """
        return self._handler.apply(config_file_path, auto_approve)



from typing import Annotated, Any, Literal

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

VARIABLES_CONFIG_V1_KEYS_ORDER = [
    "schema_version",
    "required",
    "required_sensitive",
    "overrides",
    "defaults",
]

VARIABLES_CONFIG_V1_COMMENTS: dict[str, list[str]] = {
    "required": ["  # either assign values below", "  # or run 'jd config' to use the interactive experience"],
    "required_sensitive": [
        "  # either assign values below",
        "  # or run 'jd config -s' to use the interactive experience",
    ],
    "overrides": [
        "  # set variable values as <variable-name>: <variable-value>",
        "  # delete or comment out a line to use the default",
    ],
    "defaults": [
        "  # read-only: do not modify this section",
        "  # instead add overrides in the override section",
    ],
}


class JupyterDeployVariablesConfigV1(BaseModel):
    model_config = ConfigDict(extra="allow")
    schema_version: Literal[1]
    required: dict[str, Any] = Field(default_factory=dict)
    required_sensitive: dict[str, Any] = Field(default_factory=dict)
    overrides: dict[str, Any] = Field(default_factory=dict)
    defaults: dict[str, Any] = Field(default_factory=dict)

    @field_validator("required", "required_sensitive", "overrides", "defaults", mode="before")
    @classmethod
    def ensure_dict(cls, v: Any) -> dict[str, Any]:
        return {} if v is None else v

    @model_validator(mode="after")
    def check_overrides_exist(self) -> "JupyterDeployVariablesConfigV1":
        """Ensure that all overrides have a corresponding default.

        Raises:
            ValueError if an override is not recognized.
        """
        override_var_names: set[str] = set(self.overrides.keys())
        default_var_names: set[str] = set(self.defaults.keys())

        unknown_override_names = override_var_names - default_var_names

        if unknown_override_names:
            raise ValueError(f"Unrecognized overrides: {unknown_override_names}")
        return self

    @model_validator(mode="after")
    def check_no_var_name_repeat(self) -> "JupyterDeployVariablesConfigV1":
        """Ensure that all variables are declared at most once."""
        required_var_names: set[str] = set(self.required.keys())
        sensitive_var_names: set[str] = set(self.required_sensitive.keys())
        defaults_var_names: set[str] = set(self.defaults.keys())

        req_sensitive_overlap = required_var_names.intersection(sensitive_var_names)
        req_defaults_overlap = required_var_names.intersection(defaults_var_names)
        sensitive_defaults_overlap = sensitive_var_names.intersection(defaults_var_names)

        if req_sensitive_overlap:
            raise ValueError(f"Variables definition conflict: {req_sensitive_overlap}")
        if req_defaults_overlap:
            raise ValueError(f"Variables definition conflict: {req_defaults_overlap}")
        if sensitive_defaults_overlap:
            raise ValueError(f"Variables definition conflict: {sensitive_defaults_overlap}")

        return self


# Combined type using discriminated union
JupyterDeployVariablesConfig = Annotated[JupyterDeployVariablesConfigV1, "schema_version"]
